"""This Python module defines various data object types, which are produced and processed within OVITO's data pipeline system.
It also provides the :py:class:`DataCollection` class as a container for such data objects as well as several utility classes for
computing neighbor lists and iterating over the bonds of connected to a particle.

Data containers:

  * :py:class:`DataObject` (base class for all data object types)
  * :py:class:`DataCollection` (an entire dataset made of several data objects)
  * :py:class:`PropertyContainer` (base container class storing a set of :py:class:`Property` arrays)
  * :py:class:`Particles` (specialized :py:class:`PropertyContainer` for particles)
  * :py:class:`Bonds` (specialized :py:class:`PropertyContainer` for bonds)
  * :py:class:`VoxelGrid` (specialized :py:class:`PropertyContainer` for 2d and 3d data grids)
  * :py:class:`DataTable` (specialized :py:class:`PropertyContainer` for tabulated data)

Data objects:

  * :py:class:`Property` (array of per-data-element property values)
  * :py:class:`SimulationCell` (simulation box geometry and boundary conditions)
  * :py:class:`SurfaceMesh` (polyhedral mesh representation of the boundary surface of a spatial volume)
  * :py:class:`TrajectoryLines` (set of particle trajectory lines)
  * :py:class:`TriangleMesh` (general mesh data structure made of vertices and triangular faces)
  * :py:class:`DislocationNetwork` (set of discrete dislocation lines with Burgers vector information)

Auxiliary data objects:

  * :py:class:`ElementType` (base class for element type definitions)
  * :py:class:`ParticleType` (describes a single particle or atom type)
  * :py:class:`BondType` (describes a single bond type)
  * :py:class:`DislocationSegment` (a dislocation line in a :py:class:`DislocationNetwork`)

Utility classes:

  * :py:class:`CutoffNeighborFinder` (find neighboring particles within a cutoff distance)
  * :py:class:`NearestNeighborFinder` (find *N* nearest neighbor particles)
  * :py:class:`BondsEnumerator` (efficiently iterate over the bonds connected to a particle)"""
__all__ = ['DataCollection', 'DataObject', 'TriangleMesh', 'SimulationCell', 'Property', 'PropertyContainer', 'DataTable', 'ElementType', 'SurfaceMesh', 'VoxelGrid', 'ParticleType', 'BondType', 'BondsEnumerator', 'CutoffNeighborFinder', 'NearestNeighborFinder', 'PTMNeighborFinder', 'Particles', 'Bonds', 'TrajectoryLines', 'DislocationNetwork', 'DislocationSegment']
from typing import NamedTuple, Tuple, Optional, Any, Union, Iterator, Mapping, MutableMapping, Sequence, MutableSequence, NamedTuple, TypeVar
import ovito.vis
import ovito.pipeline
import numpy
import numpy.typing
import enum
DataObjectType = TypeVar('DataObjectType', bound=DataObject)

class DataObject:
    """Abstract base class for all data object types in OVITO. 

A :py:class:`DataObject` represents a fragment of data processed in or by a data pipeline. 
See the :py:mod:`ovito.data` module for a list of different concrete data object types in OVITO. 
Data objects are typically contained in a :py:class:`DataCollection`, 
which represents a whole *data set*. Furthermore, data objects can be nested into hierarchy. 
For example, the :py:class:`Bonds` data object is part of the parent :py:class:`Particles` data object.

Data objects by themselves are non-visual objects. Visualizing the information stored in a data object in images 
is the responsibility of so-called visual elements.
A data object may be associated with a :py:class:`DataVis` element by assigning it to the data object's :py:attr:`vis` field. 
Each type of visual element exposes a set of parameters that allow you to configure the appearance of the 
data visualization in rendered images and animations. 

.. seealso:: 

    :ref:`data_model_intro` 

    :ref:`rendering_visual_elements`"""

    @property
    def identifier(self) -> str:
        """The unique identifier string of the data object. It serves as look-up key in object dictionaries, for example the :py:attr:`DataCollection.tables` collection, or as a target name in various places where a data object needs to be referenced by name, e.g. the `TimeAveragingModifier.operate_on` field. 

Data objects generated by modifiers in a pipeline typically have an automatically assigned identifier, as documented in the description of the respective modifier. When writing your own modifier function, you are responsible for giving new data objects created by your modifier function a meaningful identifier, so that subsequent modifiers in the pipeline can refer to these data objects."""
        ...

    @property
    def vis(self) -> Optional[ovito.vis.DataVis]:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    def make_mutable(self, subobj: Optional[DataObjectType]) -> Optional[DataObjectType]:
        """This helper method requests a deep copy of *subobj*, which must be a child :py:class:`DataObject` of this parent :py:class:`DataObject`. A copy will only be made in case the sub-object is currently referenced by at least one more parent object. If, however, the sub-object is exclusive owned by this :py:class:`DataObject`, no copy is made and the original sub-object is returned as is. The returned object is safe to modify without unexpected side effects, because any shared ownership is converted an exclusive ownership by the method. 

Please see the section :ref:`underscore_notation` for a discussion of object ownership and typical use-cases for this method. 

:param DataObject subobj: A existing sub-object of this parent data object, for which exclusive ownership is requested.
:return: A copy of *subobj* if its ownership was previously shared with some other parent. Otherwise the original object is returned."""
        ...

class DataCollection(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

A :py:class:`DataCollection` is a container class holding together individual *data objects*, each representing different fragments of a dataset. For example, a dataset loaded from a simulation data file may consist of particles, the simulation cell information and additional auxiliary data such as the current timestep number of the snapshots, etc. All this information is contained in one :py:class:`DataCollection`, which exposes the individual pieces of information as sub-objects, for example via the :py:attr:`DataCollection.particles`, :py:attr:`DataCollection.cell` and :py:attr:`DataCollection.attributes` fields. 

Data collections are the elementary entities that get processed within a data :py:class:`Pipeline`. Each modifier receives a data collection from the preceding modifier, alters it in some way, and passes it on to the next modifier. The output data collection of the last modifier in the pipeline is returned by the `Pipeline.compute()` method. 

A data collection essentially consists of a bunch of `DataObjects`, which are all stored in the :py:attr:`DataCollection.objects` list. Typically, you don't access the data objects through this list directly but rather use on of the special accessor fields provided by the :py:class:`DataCollection` class, which give more convenient access to data objects of a particular kind. For example, the :py:attr:`surfaces` dictionary provides key-based access to all the :py:class:`SurfaceMesh` instances currently in the data collection. 

You can programmatically add or remove data objects from a data collection by manipulating its :py:attr:`objects` list. For instance, to populate a new data collection instance that is initially empty with a new :py:class:`SimulationCell` object: 

```python
  data = DataCollection()
  cell = SimulationCell()
  data.objects.append(cell)
  assert(data.cell is cell)
```"""

    @property
    def attributes(self) -> MutableMapping[str, Union[str, int, float]]:
        """This field contains a dictionary view with all the *global attributes* currently associated with this data collection.
Global attributes are key-value pairs that represent small tokens of information, typically simple value types such as ``int``, ``float`` or ``str``. 
Every attribute has a unique identifier such as ``'Timestep'`` or ``'ConstructSurfaceMesh.surface_area'``. This identifier serves as look-up key in the :py:attr:`attributes` dictionary.
Attributes are dynamically generated by modifiers in a data pipeline or come from the data source.
For example, if the input simulation file contains timestep information, the timestep number is made available by the :py:attr:`FileSource` as the
``'Timestep'`` attribute. It can be retrieved from pipeline's output data collection:

    >>> pipeline = import_file('snapshot_140000.dump')
    >>> pipeline.compute().attributes['Timestep']
    140000

Some modifiers report their calculation results by adding new attributes to the data collection. See each modifier's
reference documentation for the list of attributes it generates. For example, the number of clusters identified by the
:py:class:`ClusterAnalysisModifier` is available in the pipeline output as an attribute named
``ClusterAnalysis.cluster_count``::

    pipeline.modifiers.append(ClusterAnalysisModifier(cutoff = 3.1))
    data = pipeline.compute()
    nclusters = data.attributes["ClusterAnalysis.cluster_count"]

The :py:func:`ovito.io.export_file` function can be used to output dynamically computed attributes to a text file, possibly as functions of time::

    export_file(pipeline, "data.txt", "txt/attr",
        columns = ["Timestep", "ClusterAnalysis.cluster_count"],
        multiple_frames = True)

If you are writing your own modifier function, you let it add new attributes to a data collection.
In the following example, the :py:class:`CommonNeighborAnalysisModifier` first inserted into the
pipeline generates the ``'CommonNeighborAnalysis.counts.FCC'`` attribute to report the number of atoms that
have an FCC-like coordination. To compute an atomic *fraction* from that, we need to divide the count by the total number of
atoms in the system. To this end, we append a user-defined modifier function
to the pipeline, which computes the fraction and outputs the value as a new attribute named ``'fcc_fraction'``.

```python
  pipeline.modifiers.append(CommonNeighborAnalysisModifier())
              
  def compute_fcc_fraction(frame, data):
      n_fcc = data.attributes['CommonNeighborAnalysis.counts.FCC']
      data.attributes['fcc_fraction'] = n_fcc / data.particles.count
  
  pipeline.modifiers.append(compute_fcc_fraction)
  print(pipeline.compute().attributes['fcc_fraction'])
```"""
        ...

    @property
    def cell(self) -> SimulationCell:
        """Returns the :py:class:`SimulationCell` object, which stores the cell vectors and periodic boundary
condition flags, or ``None`` if there is no cell information associated with this data collection. 

Note: The :py:class:`SimulationCell` data object may be read-only if it is currently shared by
several data collections. Use the :py:attr:`cell_` field instead to request a mutable cell object 
if you intend to modify it."""
        ...

    @property
    def cell_(self) -> SimulationCell:
        ...

    @property
    def dislocations(self) -> DislocationNetwork:
        """Returns the :py:class:`DislocationNetwork` data object from this data collection or ``None`` if there isn't one."""
        ...

    @property
    def particles(self) -> Particles:
        """Returns the :py:class:`Particles` object from this data collection, which stores the particle properties and -as a sub-object- the 
:py:attr:`~Particles.bonds` between particles. ``None`` is returned if the data collection does not contain any particle data.

Note that the :py:class:`Particles` object may be marked as read-only if it is currently shared by several data collections.
If you intend to modify the particles container or its sub-objects in any way, e.g. add, remove or modify particle properties, 
you must request a mutable version of the particles object using the :py:attr:`particles_` accessor instead."""
        ...

    @property
    def particles_(self) -> Particles:
        ...

    @property
    def trajectories(self) -> TrajectoryLines:
        """Returns the :py:class:`TrajectoryLines` object, which holds the continuous particle trajectories traced 
by the :py:class:`GenerateTrajectoryLinesModifier`. 
``None`` is returned if the data collection does not contain a :py:class:`TrajectoryLines` object."""
        ...

    @property
    def objects(self) -> MutableSequence[DataObject]:
        """The unordered list of all `DataObjects` stored in this data collection. You can add or remove data objects in this list as needed. 

Note that typically you don't have to work with this list directly, because the :py:class:`DataCollection` class provides several convenience accessor fields for the different flavors of objects in this mixed list. For example, the :py:attr:`DataCollection.particles` field returns the :py:class:`Particles` object from this data objects list. Also, dictionary views such as :py:attr:`DataCollection.tables` provide key-based access to a particular class of data objects from this list."""
        ...

    @property
    def grids(self) -> Mapping[str, VoxelGrid]:
        """Returns a dictionary view providing key-based access to all `VoxelGrids` in 
this data collection. Each :py:class:`VoxelGrid` has a unique :py:attr:`identifier` key, 
which you can use to look it up in this dictionary. To find out which voxel grids exist in the data collection and what 
their identifier keys are, use

```python
  print(data.grids)
```

Then retrieve the desired :py:class:`VoxelGrid` from the collection using its identifier key, e.g.

```python
  charge_density_grid = data.grids['charge-density']
  print(charge_density_grid.shape)
```"""
        ...

    @property
    def surfaces(self) -> Mapping[str, SurfaceMesh]:
        """Returns a dictionary view providing key-based access to all :py:class:`SurfaceMesh` objects in 
this data collection. Each :py:class:`SurfaceMesh` has a unique :py:attr:`identifier` key, 
which can be used to look it up in the dictionary. 
See the documentation of the modifier producing the surface mesh to find out what the right key is, or use

```python
  print(data.surfaces)
```

to see which identifier keys exist. Then retrieve the desired :py:class:`SurfaceMesh` object from the collection using its identifier 
key, e.g.

```python
  surface = data.surfaces['surface']
  print(surface.get_vertices())
```"""
        ...

    @property
    def tables(self) -> Mapping[str, DataTable]:
        """A dictionary view of all :py:class:`DataTable` objects in
this data collection. Each :py:class:`DataTable` has a unique :py:attr:`identifier` key, 
which can be used to look it up in this dictionary. You can use

```python
  print(data.tables)
```

to find out which table identifiers exist in the dictionary. Modifiers that generate a data table typically 
assign a predefined identifier, which can be found in their documentation.
Use the key string to retrieve the desired :py:class:`DataTable` from the dictionary, e.g.

```python
  rdf = data.tables['coordination-rdf']
  print(rdf.xy())
```"""
        ...

    def apply(self, modifier: ovito.pipeline.Modifier, frame: Optional[int]=None) -> None:
        """This method applies a :py:class:`Modifier` function to the data in this collection in place.

It allows modifying a data collection with one of Ovito's built-in modifiers directly without the need to build up a complete
:py:class:`Pipeline` first. In contrast to a data pipeline, the :py:meth:`!apply()` method
executes the modifier function immediately and changes the input data in place. In other words, the original data in this :py:class:`DataCollection`
will be replaced by the output produced by the invoked modifier function. Note that it is possible to create a copy of
the original data collection using the :py:meth:`.clone` method if needed. The following code example
demonstrates how to use the :py:meth:`!apply()` method to successively modify the state of a dataset:

```python
  from ovito.io import import_file
  from ovito.modifiers import *
  
  data = import_file("input/simulation.dump").compute()
  data.apply(CoordinationAnalysisModifier(cutoff=2.9))
  data.apply(ExpressionSelectionModifier(expression="Coordination<9"))
  data.apply(DeleteSelectedModifier())
```

Note that it is typically possible to achieve the same result by first populating a :py:class:`Pipeline` with the modifiers and then calling its
:py:meth:`compute` method at the very end:

```python
  pipeline = import_file("input/simulation.dump")
  pipeline.modifiers.append(CoordinationAnalysisModifier(cutoff=2.9))
  pipeline.modifiers.append(ExpressionSelectionModifier(expression="Coordination<9"))
  pipeline.modifiers.append(DeleteSelectedModifier())
  data = pipeline.compute()
```

Also note that :py:meth:`!apply()` may be called from a user-defined modifier function (see :py:class:`PythonScriptModifier`)
in order to invoke a built-in modifier as a sub-routine:

```python
  # A user-defined modifier function that calls the built-in ColorCodingModifier
  # as a sub-routine to assign a color to each atom based on some property
  # created within the function itself:
  def modify(frame, data):
      data.particles_.create_property('idx', data=numpy.arange(data.particles.count))
      data.apply(ColorCodingModifier(property='idx'), frame)
  
  # Set up a data pipeline that uses the user-defined modifier function:
  pipeline = import_file("input/simulation.dump")
  pipeline.modifiers.append(modify)
  data = pipeline.compute()
```

:param modifier: The :py:class:`Modifier` that will be called by the method to alter the data in place.
:param int frame: An optional animation frame number that will be passed to the modifier function, which may use it to implement time-dependent modifications."""
        ...

    def clone(self) -> DataCollection:
        """Returns a copy of this :py:class:`DataCollection` containing the same data objects as the original.

The method may be used to retain a copy of the original data before modifying a data collection in place,
for example using the :py:meth:`.apply` method:

```python
  original = data.clone()
  data.apply(ExpressionSelectionModifier(expression="Position.Z < 0"))
  data.apply(DeleteSelectedModifier())
  print("Number of atoms before:", original.particles.count)
  print("Number of atoms after:", data.particles.count)
```

Note that the :py:meth:`!clone()` method performs an inexpensive, shallow copy, meaning that the newly created collection will still share
the data objects with the original collection. Data objects that are shared by two or more data collections are
protected against modification by default to avoid unwanted side effects.
Thus, in order to subsequently modify the data objects in either the original collection or its
copy, you will have to use the underscore notation or the :py:meth:`DataObject.make_mutable` method to explicitly
request a deep copy of the particular data object(s) you want to modify. For example:

```python
  copy = data.clone()
  # Data objects are shared by original and copy:
  assert(copy.cell is data.cell)
  
  # In order to modify the SimulationCell in the dataset copy, we must request
  # a mutable version of the SimulationCell using the 'cell_' accessor:
  copy.cell_.pbc = (False, False, False)
  
  # As a result, the cell object in the second data collection has been replaced
  # with a deep copy and the two data collections no longer share the same
  # simulation cell object:
  assert(copy.cell is not data.cell)
```"""
        ...

class ElementType(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

This class represents a single type of data elements, for example a particular atom or bond type. 
It serves as common base class for the :py:class:`ParticleType` and :py:class:`BondType` classes, 
which represent these more specific types. 

Each type has a unique numeric :py:attr:`.id`, which is used when looking up types given some 
numeric value in a *typed* :py:class:`Property` array. An example for a typed property is the particle property ``'Particle Type'``. 
The property object manages all its associated types in its `Property.types` list. 

The `Property.type_by_id()` and `Property.type_by_name()` 
methods allow looking up a certain :py:class:`ElementType` based on its numeric identifier or name string."""

    @property
    def color(self) -> ovito.vis.Color:
        """The color used when rendering elements of this type. This is a RGB tuple with components in the range 0.0 -- 1.0.

:Default: ``(1.0, 1.0, 1.0)``"""
        ...

    @property
    def enabled(self) -> bool:
        """Controls whether this type is currently *active* or *inactive*. This flag currently has a meaning only in the context of atomic structure identification. Some analysis modifiers manage a list of the structure types they can identify (e.g. FCC, BCC, etc.). The identification of individual structure types can be turned on or off by the user by changing their :py:attr:`enabled` flag. See `StructureIdentificationModifier.structures` for further information. 

:Default: ``True``"""
        ...

    @property
    def id(self) -> int:
        """The unique numeric identifier of the type (typically some positive ``int``). The identifier is and must be unique among all element types in the :py:attr:`types` list of a typed :py:class:`Property`. Thus, if you create a new element type, make sure you give it a unique id before inserting it into the :py:attr:`types` list of a typed property. 

:Default: ``0``"""
        ...

    @property
    def name(self) -> str:
        """The name of this type, e.g. the chemical element symbol of an atom type. This string may be empty, in which case its numeric :py:attr:`id` is the only way of referring to this type. 

:Default: ``''``"""
        ...

class ParticleType(ElementType):
    """Base: :py:class:`ovito.data.ElementType`

This data object describes one particle or atom type. In atomistic simulations, each chemical element is typically represented by an instance of
the :py:class:`ParticleType` class. The property fields of the class control how the particles of that type get visualized in terms of e.g. color, particle radius, shape, etc.

The :py:class:`ParticleType` class inherits several general data fields from its base class :py:class:`ElementType`, e.g. the :py:attr:`~ElementType.color`, :py:attr:`~ElementType.name`
and :py:attr:`~ElementType.id` fields. It adds specific fields for particles: :py:attr:`radius` and :py:attr:`shape`. Furthermore, the class has additional
fields controlling the visual appearance of particles with user-defined shapes.

The :py:class:`ParticleType` instances are all stored in the :py:class:`Property` object with the name ``'Particle Type'``, which also stores for each particle what
its type is. The association between particles and particle types is established via a unique numeric type :py:attr:`~ElementType.id`. The following code shows how to iterate over
all particle types in a dataset, which are listed in the :py:attr:`Property.types` field of the :py:attr:`particle_types` property:

```python
  # Access the property with the name 'Particle Type':
  prop = data.particles.particle_types
  
  # Print list of particle types (their numeric IDs and names)
  for t in prop.types: print(f'ID {t.id} -> {t.name}')
  
  # Print the numeric type ID of each particle:
  for tid in prop[...]: print(tid)
```

The order in which the particle types are stored in the :py:attr:`Property.types` list is arbitrary, and the unique numeric IDs of particle types have no specific meaning in general.
A common operation is to find the :py:class:`~ParticleType` in the list corresponding to a given numeric ID. For this look up operation, the :py:class:`Property` class provides the :py:meth:`type_by_id` method:

```python
  # Look up the particle type with unique ID 2:
  t = prop.type_by_id(2)
  print(t.name, t.color, t.radius)
  
  # Iterate over all particles and print their type's name:
  for tid in prop[...]:
      print(prop.type_by_id(tid).name)
```

Another common operation is to look up a particle type by name, for example the type representing a certain chemical element. 
For this kind of look up operation, the :py:meth:`type_by_name` method may be used, which assumes that each type has a unique :py:attr:`name` (which may not always be true):

```python
  # Print numeric ID of particle type 'Si':
  print(prop.type_by_name('Si').id)
```"""

    @property
    def backface_culling(self) -> bool:
        """Activates back-face culling for the user-defined particle shape mesh to speed up rendering. If turned on, polygonal sides of the shape mesh facing away from the viewer will not be rendered. You can turn this option off if the particle's shape is not closed and two-sided rendering is required. This option only has an effect if a user-defined shape has been assigned to the particle type using the :py:meth:`.load_shape` method. 

:Default: ``True``"""
        ...

    @property
    def highlight_edges(self) -> bool:
        """Activates the highlighting of the polygonal edges of the user-defined particle shape during rendering. This option only has an effect if a user-defined shape has been assigned to the particle type using the :py:meth:`.load_shape` method. 

:Default: ``False``"""
        ...

    @property
    def mass(self) -> float:
        """The mass of this particle type. 

:Default: ``0.0``"""
        ...

    @property
    def radius(self) -> float:
        """This property controls the display radius of the particles of this type. 

When set to zero, particles of this type will be rendered using the standard size specified by the `ParticlesVis.radius` parameter. Furthermore, precedence is given to any per-particle sizes assigned to the ``Radius`` particle property if that property has been defined. 

:Default: ``0.0``


The following example script demonstrates how to set the display radii of two particle types loaded from a simulation file, which can be accessed through the `Property.types` list of the ``Particle Type`` particle property. 

```python
  pipeline = import_file("input/simulation.dump")
  pipeline.add_to_scene()
  
  def setup_particle_types(frame, data):
      types = data.particles_.particle_types_
      types.type_by_id_(1).name = "Cu"
      types.type_by_id_(1).radius = 1.35
      types.type_by_id_(2).name = "Zr"
      types.type_by_id_(2).radius = 1.55
  pipeline.modifiers.append(setup_particle_types)
```"""
        ...

    @property
    def shape(self) -> ovito.vis.ParticlesVis.Shape:
        """Selects the geometric shape used when rendering particles of this type. Supported modes are:

   * ``ParticlesVis.Shape.Unspecified`` (default) 
   * ``ParticlesVis.Shape.Sphere``
   * ``ParticlesVis.Shape.Box``
   * ``ParticlesVis.Shape.Circle``
   * ``ParticlesVis.Shape.Square``
   * ``ParticlesVis.Shape.Cylinder``
   * ``ParticlesVis.Shape.Spherocylinder``
   * ``ParticlesVis.Shape.Mesh``


By default, the standard particle shape that is set in the `ParticlesVis` visual element is used to render particles of this type. Parameter values other than ``Unspecified`` allow you to control the rendering shape on a per-type basis. Mode ``Sphere`` includes ellipsoid and superquadric particle shapes, which are enabled by the presence of the ``Aspherical Shape`` and ``Superquadric Roundness`` particle properties. 

The :py:meth:`load_shape` method lets you specify a user-defined mesh geometry for this particle type. Calling this method automatically switches the shape parameter to mode ``Mesh``. 

Setting the shapes of particle types permanently, i.e., for all frames of a loaded simulation trajectory, typically requires a user-defined modifier function. This function is inserted into the :py:class:`Pipeline` to make the necessary changes to the :py:class:`ParticleType` objects associated with the :py:class:`Property` named ``Particle Type``: 

```python
  from ovito.io import import_file
  from ovito.vis import *
  
  # Load a simulation file containing numeric particle types 1, 2, 3, ...
  pipeline = import_file("input/nylon.data")
  pipeline.add_to_scene()
  
  # Set the default particle shape in the ParticlesVis visual element, 
  # which will be used by all particle types for which we do not specify a different shape below.
  pipeline.compute().particles.vis.shape = ParticlesVis.Shape.Box
  pipeline.compute().particles.vis.radius = 1.0
  
  # A user-defined modifier function that configures the shapes of particle types 1 and 2:
  def setup_particle_types(frame, data): 
      # Write access to property 'Particle Type':
      types = data.particles_.particle_types_  
      # Write access to numeric ParticleTypes, which are sub-objects of the Property object:
      types.type_by_id_(1).radius = 0.5
      types.type_by_id_(1).shape = ParticlesVis.Shape.Cylinder
      types.type_by_id_(2).radius = 1.2
      types.type_by_id_(2).shape = ParticlesVis.Shape.Sphere
  pipeline.modifiers.append(setup_particle_types)
  
  # Render a picture of the 3d scene:
  vp = Viewport(camera_dir = (-2,1,-1))
  vp.zoom_all()
  vp.render_image(filename='output/particles.png', size=(320,240), renderer=TachyonRenderer())
```"""
        ...

    @property
    def use_mesh_color(self) -> bool:
        """Use the intrinsic mesh color(s) instead of the particle color when rendering particles of this type. This option only has an effect if a user-defined shape has been assigned to the particle type using the :py:meth:`.load_shape` method. 

:Default: ``False``"""
        ...

    @property
    def vdw_radius(self) -> float:
        """The van der Waals radius of the particle type. This value is used by the :py:class:`CreateBondsModifier` to decide which pairs of particles are close enough to be connected by a bond. In contrast to the :py:attr:`radius` parameter, the van der Waals radius does not affect the visual appearance of the particles of this type. 

:Default: ``0.0``"""
        ...

    def load_defaults(self) -> None:
        """Given the type's chemical :py:attr:`name`, which must have been set before calling this method, initializes the type's :py:attr:`color` and :py:attr:`radius` fields with default values from OVITO's internal database of chemical elements. This method is useful when creating new atom types while building up a molecule structure."""
        ...

    def load_shape(self, filepath: str) -> None:
        """Assigns a user-defined shape to the particle type. Particles of this type will subsequently be rendered using the polyhedral mesh loaded from the given file. The method will automatically detect the format of the geometry file and supports standard file formats such as OBJ, STL and VTK that contain triangle meshes, see this table. 

The shape loaded from the geometry file will be scaled with the :py:attr:`radius` value set for this particle type or the per-particle value stored in the ``Radius`` particle property if present. The shape of each particle will be rendered such that its origin is located at the coordinates of the particle. 

The following example script demonstrates how to load a user-defined shape for the first particle type (index 0) loaded from a LAMMPS dump file, which can be accessed through the `Property.types` list of the ``Particle Type`` particle property. 

```python
  pipeline = import_file("input/simulation.dump")
  pipeline.add_to_scene()
  
  types = pipeline.source.data.particles_.particle_types_
  types.type_by_id_(1).load_shape("input/tetrahedron.vtk")
  types.type_by_id_(1).highlight_edges = True
```"""
        ...

class BondType(ElementType):
    """Base: :py:class:`ovito.data.ElementType`

Represents a bond type. This class inherits all its fields from the :py:class:`ElementType` base class. 

You can enumerate the list of defined bond types by accessing the `bond_types` bond property object: 

```python
  bond_type_list = data.particles.bonds.bond_types.types
  for bond_type in bond_type_list:
      print(bond_type.id, bond_type.name, bond_type.color)
```"""
    pass

class DislocationSegment:
    """A single dislocation line from a :py:class:`DislocationNetwork`. 

The list of dislocation segments is returned by the :py:attr:`DislocationNetwork.segments` attribute."""

    @property
    def cluster_id(self) -> int:
        """The numeric identifier of the crystal cluster of atoms containing this dislocation segment. 

The true Burgers vector of the segment is expressed in the local coordinate system of this crystal cluster."""
        ...

    @property
    def custom_color(self) -> ovito.vis.Color:
        """The RGB color value to be used for visualizing this particular dislocation line, overriding the default coloring scheme imposed by the `DislocationVis.coloring_mode` setting. The custom color is only used if its RGB components are non-negative (i.e. in the range 0-1); otherwise the line will be rendered using the computed color depending on the line's Burgers vector. 

:Default: ``(-1.0, -1.0, -1.0)``"""
        ...

    @property
    def id(self) -> int:
        """The unique identifier of this dislocation segment."""
        ...

    @property
    def is_infinite_line(self) -> bool:
        """This property indicates whether this segment is an infinite line passing through a periodic simulation box boundary. A segment is considered infinite if it is a closed loop and its start and end points do not coincide. 

See also the :py:attr:`is_loop` property."""
        ...

    @property
    def is_loop(self) -> bool:
        """This property indicates whether this segment forms a closed dislocation loop. Note that an infinite dislocation line passing through a periodic boundary is also considered a loop. 

See also the :py:attr:`is_infinite_line` property."""
        ...

    @property
    def length(self) -> float:
        """Returns the length of this dislocation segment."""
        ...

    @property
    def points(self) -> numpy.typing.ArrayLike:
        """The list of space points that define the shape of this dislocation segment. This is a *N* x 3 Numpy array, where *N* is the number of points along the segment. For closed loops, the first and the last point coincide."""
        ...

    @property
    def spatial_burgers_vector(self) -> ovito.vis.Vector3:
        """The Burgers vector of the segment, expressed in the global coordinate system of the simulation. This vector is calculated by transforming the true Burgers vector from the local lattice coordinate system to the global simulation coordinate system using the average orientation matrix of the crystal cluster the dislocation segment is embedded in."""
        ...

    @property
    def true_burgers_vector(self) -> ovito.vis.Vector3:
        """The Burgers vector of the segment, expressed in the local coordinate system of the crystal. Also known as the True Burgers vector."""
        ...

class Property(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

Stores the property values for an array of data elements (e.g. particles, bonds or voxels). 

Each property, for example a particle property, is represented by one :py:class:`Property` object storing the property values for all particles. 
Thus, a :py:class:`Property` object is basically an array of values whose length matches the number of data elements. 

All :py:class:`Property` objects belonging to the same class of data elements, for example all particle properties, are managed by 
a :py:class:`PropertyContainer`. In the case of particle properties, the corresponding container class is the 
:py:class:`Particles` class, which is a specialization of the generic :py:class:`PropertyContainer` base class.

Data access

A :py:class:`Property` object behaves like a Numpy array. For example, you can access the property value for the *i*-th data element using indexing:: 

     positions = data.particles['Position']
     print('Position of first particle:', positions[0])
     print('Z-coordinate of second particle:', positions[1,2])
     for xyz in positions: 
         print(xyz)

Element indices start at zero. Properties can be either vectorial (e.g. velocity vectors are stored as an *N* x 3 array) 
or scalar (1-d array of length *N*). The length of the first array dimension is in both cases equal to 
the number of data elements (number of particles in the example above). Array elements can either be of data type ``float64``, ``int32`` or ``int64``. 

If you want to modify the per-element values in a property array, make sure you are working with a modifiable version of the array 
by employing the :pyunderscore notation, e.g.:: 

    modifiable_positions = data.particles_['Position_']
    modifiable_positions[0] = (0.1, 2.3, 0.5)

Typed properties

The standard particle property ``'Particle Type'`` stores the types of particles encoded as integer values, e.g.: 

    >>> data = node.compute()
    >>> tprop = data.particles['Particle Type']
    >>> print(tprop[...])
    [2 1 3 ..., 2 1 2]

Here, each number in the property array refers to one of the particle types (e.g. 1=Cu, 2=Ni, 3=Fe, etc.). The defined particle types, each one represented by 
an instance of the :py:class:`ParticleType` auxiliary class, are stored in the :py:attr:`types` array 
of the :py:class:`Property`. Each type has a unique :py:attr:`~ElementType.id`, a human-readable :py:attr:`~ElementType.name` 
and other attributes like :py:attr:`~ElementType.color` and :py:attr:`~ParticleType.radius` that control the 
visual appearance of particles belonging to the type:

    >>> for type in tprop.types:
    ...     print(type.id, type.name, type.color, type.radius)
    ...
    1 Cu (0.188 0.313 0.972) 0.74
    2 Ni (0.564 0.564 0.564) 0.77
    3 Fe (1 0.050 0.050) 0.74

IDs of types typically start at 1 and form a consecutive sequence as in the example above. 
Note, however, that the :py:attr:`types` list may store the :py:class:`ParticleType` objects in an arbitrary order. 
Thus, in general, it is not valid to directly use a type ID as an index into the :py:attr:`types` array. 
Instead, the :py:meth:`.type_by_id` method should be used to look up the :py:class:`ParticleType`:: 

    >>> for i,t in enumerate(tprop): # (loop over the type ID of each particle)
    ...     print('Atom', i, 'is of type', tprop.type_by_id(t).name)
    ...
    Atom 0 is of type Ni
    Atom 1 is of type Cu
    Atom 2 is of type Fe
    Atom 3 is of type Cu

Similarly, a :py:meth:`.type_by_name` method exists that looks up a :py:attr:`ParticleType` by name. 
For example, to count the number of Fe atoms in a system:

    >>> Fe_type_id = tprop.type_by_name('Fe').id   # Determine ID of the 'Fe' type
    >>> numpy.count_nonzero(tprop == Fe_type_id)   # Count particles having that type ID
    957

Note that OVITO supports multiple type classifications. For example, in addition to the ``'Particle Type'`` standard particle property, 
which stores the chemical types of atoms (e.g. C, H, Fe, ...), the ``'Structure Type'`` property may hold the structural types computed for atoms 
(e.g. FCC, BCC, ...) maintaining its own list of known structure types in the :py:attr:`types` array."""

    @property
    def component_count(self) -> int:
        """The number of vector components if this is a vector property; or 1 if this is a scalar property."""
        ...

    @property
    def component_names(self) -> Tuple[str]:
        """The list of names of the vector components if this is a vector property. For example, for the ``Position`` particle property this field contains ``['X', 'Y', 'Z']``."""
        ...

    @property
    def name(self) -> str:
        """The name of the property."""
        ...

    @property
    def types(self) -> MutableSequence[ElementType]:
        """The list of :py:class:`ElementType` instances attached to this property. 

Note that the element types may be stored in arbitrary order in this list. Thus, it is not valid to use a numeric type ID as an index into this list."""
        ...

    def type_by_id(self, id: int, raise_error: bool=True) -> Optional[ElementType]:
        """Looks up and returns the :py:class:`ElementType` with the given unique numeric ID in this property's :py:attr:`types` list.
Depending on *raise_error*, raises a ``KeyError`` or returns ``None`` if no type with the numeric ID exists.


Usage example:

```python
  # Iterate over the numeric per-particle types stored in the 'Structure Type' 
  # particle property array and print the corresponding human-readable type names:
  for index, type_id in enumerate(data.particles.structure_types):
      print("Atom {} is a {} atom".format(
          index, data.particles.structure_types.type_by_id(type_id).name))
```

An :ref:`"underscore" version <underscore_notation>` of the method exists, which should be used whenever you intend to modify the returned type object. :py:meth:`!type_by_id_()` implicitly calls :py:meth:`make_mutable` on the :py:class:`ElementType` to make sure it can be changed without unexpected side effects:

```python
  # Give the numeric atom types from a LAMMPS simulation some names:
  data.particles_.particle_types_.type_by_id_(1).name = 'C' 
  data.particles_.particle_types_.type_by_id_(2).name = 'H' 
```"""
        ...

    def type_by_id_(self, id: int, raise_error: bool=True) -> Optional[ElementType]:
        ...

    def type_by_name(self, name: str, raise_error: bool=True) -> Optional[ElementType]:
        """Looks up and returns the :py:class:`ElementType` with the given name in this property's :py:attr:`types` list.
If multiple types exists with the same name, the first type is returned. 
Depending on *raise_error*, raises a ``KeyError`` or returns ``None`` if there isn't a type with that name. 


Usage example:

```python
  # Look up the numeric ID of atom type Si and count how many times it appears in the 'Particle Type' array
  id_Si = data.particles.particle_types.type_by_name('Si').id
  Si_atom_count = numpy.count_nonzero(data.particles.particle_types == id_Si)
```

An :ref:`"underscore" version <underscore_notation>` of the method exists, which should be used whenever you intend to modify the returned type object. :py:meth:`!type_by_name_()` implicitly calls :py:meth:`make_mutable` on the :py:class:`ElementType` to make sure it can be changed without unexpected side effects:

```python
  # Rename a structure type created by the PTM modifier:
  data.particles_.structure_types_.type_by_name_('Hexagonal diamond').name = 'Wurtzite'
```"""
        ...

    def type_by_name_(self, name: str, raise_error: bool=True) -> Optional[ElementType]:
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[numpy.typing.NDArray]:
        ...

    def __getitem__(self, idx: Any) -> numpy.typing.NDArray:
        ...

class PropertyContainer(DataObject, Mapping[str, Property]):
    """Base: :py:class:`ovito.data.DataObject`

A dictionary-like object storing a set of :py:class:`Property` objects.

It implements the ``collections.abc.Mapping`` interface. That means it can be used 
like a standard read-only Python ``dict`` object to access the properties by name, e.g.: 

```python
  data = pipeline.compute()
  
  positions = data.particles['Position']
  has_selection = 'Selection' in data.particles
  name_list = data.particles.keys()
```

New properties are typically added to a container with a call to :py:meth:`.create_property` as described here. 
To remove an existing property from a container, you can use Python's ``del`` statement: 

```python
  del data.particles_['Selection']
```

OVITO has several concrete implementations of the abstract :py:class:`PropertyContainer` interface: 

    * :py:class:`Particles`
    * :py:class:`Bonds`
    * :py:class:`VoxelGrid`
    * :py:class:`DataTable`
    * :py:class:`TrajectoryLines`
    * :py:attr:`SurfaceMesh.vertices`
    * :py:attr:`SurfaceMesh.faces`
    * :py:attr:`SurfaceMesh.regions`"""

    @property
    def count(self) -> int:
        """The number of data elements in this container, e.g. the number of particles. This value is always equal to the lengths of the :py:class:`Property` arrays managed by this container."""
        ...

    @property
    def title(self) -> str:
        """The title of the data object under which it appears in the user interface of OVITO."""
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[str]:
        ...

    def __getitem__(self, key: str) -> Property:
        ...

    def create_property(self, name: str, dtype: Optional[numpy.typing.DTypeLike]=None, components: Optional[int]=None, data: Optional[numpy.typing.ArrayLike]=None) -> ovito.data.Property:
        """Adds a new property to the container and optionally initializes it with
the per-element data provided by the *data* parameter. The method returns the new :py:class:`Property`
instance.

The method allows to create *standard* as well as *user-defined* properties.
To create a *standard* property, one of the standard property names must be provided as *name* argument:

```python
  colors = numpy.random.random_sample(size = (data.particles.count, 3))
  data.particles_.create_property('Color', data=colors)
```

The length of the provided *data* array must match the number of existing elements in the container, which is given by the :py:attr:`count` attribute.
You can alternatively assign the per-element values to the property after its construction:

```python
  prop = data.particles_.create_property('Color')
  prop[...] = numpy.random.random_sample(size = prop.shape)
```

To create a *user-defined* property, use a non-standard property name:

```python
  values = numpy.arange(0, data.particles.count, dtype=int)
  data.particles_.create_property('myint', data=values)
```

In this case the data type and the number of vector components of the new property are inferred from
the provided *data* Numpy array. Providing a one-dimensional array creates a scalar property while
a two-dimensional array creates a vectorial property.
Alternatively, the *dtype* and *components* parameters can be specified explicitly
if you are going to assign the property values at a later time:

```python
  prop = data.particles_.create_property('myvector', dtype=float, components=3)
  prop[...] = numpy.random.random_sample(size = prop.shape)
```

If the property to be created already exists in the container, it is replaced with a new one.
The existing per-element data from the old property is however retained if *data* is ``None``.

Note: If the container contains no properties yet, then the number of elements (e.g. particles or bonds) is still undefined.
In this case the :py:meth:`create_property` method lets you *define* the number of elements when inserting the very first property
by specifying a *data* array of the desired length. For example, to create a new :py:class:`Particles` container from scratch
with 10 particles, a Numpy array of length 10 is used to initialize the ``Position`` particle property:

```python
  # An empty Particles container to begin with:
  particles = Particles()
  
  # Create 10 particles with random xyz coordinates:
  xyz = numpy.random.random_sample(size = (10,3))
  particles.create_property('Position', data=xyz)
```

After the initial ``Positions`` property has been created, the number of particles in the container is now determined and any
subsequently added properties must have the exact same length.

:param name: Either a standard property type constant or a name string.
:param data: An optional data array with per-element values for initializing the new property.
                The size of the array must match the element :py:attr:`count` of the container
                and the shape must be consistent with the number of components of the property to be created.
:param dtype: The element data type when creating a user-defined property. Must be either ``int`` or ``float``.
:param int components: The number of vector components when creating a user-defined property.
:returns: The newly created :py:class:`Property` object."""
        ...

    def delete_elements(self, mask: numpy.typing.ArrayLike) -> None:
        """Deletes a subset of the elements from this container. The elements to be deleted must be specified in terms
of a 1-dimensional mask array having the same length as the container (see :py:attr:`count`).
The method will delete those elements whose corresponding mask value is non-zero, i.e., the ``i``-th element 
will be deleted if ``mask[i]!=0``.

For example, to delete all currently selected particles, i.e., the subset of particles whose ``Selection`` property 
is non-zero, one would simply write:

```python
  data.particles_.delete_elements(data.particles['Selection'])
```

The effect of this statement is the same as for applying the :py:class:`DeleteSelectedModifier` to the particles list."""
        ...

    def delete_indices(self, indices: numpy.typing.ArrayLike) -> None:
        """Deletes a subset of the elements from this container. The elements to be deleted must be specified in terms
of a sequence of indices, all in the range 0 to :py:attr:`count`-1. The method accepts any type of iterable object,
including sequence types and generators. 

For example, to delete every other particle, one could use Python's ``range()`` function to generate
all even indices up to the length of the particle container:

```python
  data.particles_.delete_indices(range(0, data.particles.count, 2))
```"""
        ...

class SimulationCell(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

This object stores the geometric shape and boundary conditions of the simulation box.
Typically there is exactly one :py:class:`SimulationCell` object in a :py:class:`DataCollection`, which is accessible through 
the :py:attr:`~DataCollection.cell` field:

```python
  data = pipeline.compute()
  print(data.cell[...])   # Use [...] to cast SimulationCell object to a NumPy array
```

The cell matrix

The geometry of the simulation cell is encoded as a 3x4 matrix :math:`\\mathbf{M}`. 
The first three columns :math:`\\mathbf{a}`, :math:`\\mathbf{b}`, :math:`\\mathbf{c}` of the matrix are the vectors spanning the three-dimensional parallelepiped
in Cartesian space. The fourth column specifies the Cartesian coordinates of the cell's origin :math:`\\mathbf{o}` within the global
simulation coordinate system:

.. math::

   \\mathbf{M} = \\begin{pmatrix}
      a_x & b_x & c_x & o_x \\\\
      a_y & b_y & c_y & o_y \\\\
      a_z & b_z & c_z & o_z \\\\
      \\end{pmatrix}

The cell matrix is represented by a two-dimensional NumPy array of shape (3,4) using row-major storage order:

```python
  a = data.cell[:,0]
  b = data.cell[:,1]
  c = data.cell[:,2]
  o = data.cell[:,3]
```

The :py:attr:`is2D` flag of the simulation cell indicates whether the system is two-dimensional. 
The cell matrix of a 2d system also has the 3x4 shape, but the cell vector :math:`\\mathbf{c}` and the 
last row of the cell matrix are ignored by many computations in OVITO if the system is marked as 2d.

Periodic boundary conditions

The :py:attr:`pbc` field stores a tuple of three Boolean flags that indicate for each 
cell vector whether the system is periodic in that direction or not. OVITO uses that information
in various computations. If the system is two-dimensional, the value of the third pbc flag
is ignored.

Modifying the simulation cell

When you modify the entries of the cell matrix, make sure you use the :pyunderscore notation 
to request a modifiable version of the :py:class:`SimulationCell` object: 

```python
  # Make cell twice as large along the Y direction by scaling the second cell vector:
  data.cell_[:,1] *= 2.0
```

Reset the simulation cell to an orthogonal box :math:`(L_x, L_y, L_z)` centered at the origin:

```python
  lx = 20.0; ly = 10.0; lz = 8.0
  data.cell_[:,0] = (lx, 0, 0)
  data.cell_[:,1] = (0, ly, 0)
  data.cell_[:,2] = (0, 0, lz)
  data.cell_[:,3] = numpy.dot((-0.5, -0.5, -0.5), data.cell[:3,:3])
  data.cell_.pbc = (True, True, True)
```

Visual representation

Every :py:class:`SimulationCell` has an attached :py:class:`SimulationCellVis` element,
which controls the visual appearance of the wireframe box in rendered images. It can be accessed via the
:py:attr:`~DataObject.vis` attribute inherited from the :py:class:`DataObject` base class:

```python
  # Change display color of simulation cell to red:
  data.cell.vis.rendering_color = (1.0, 0.0, 0.0)
```"""

    @property
    def is2D(self) -> bool:
        """Specifies whether the system is two-dimensional (instead of three-dimensional). For two-dimensional systems, the third :py:attr:`pbc` flag and the cell vector :math:`\\mathbf{c}` are typically ignored. 

:Default: ``False``"""
        ...

    @property
    def pbc(self) -> Tuple[bool, bool, bool]:
        """A tuple of three Boolean flags specifying whether periodic boundary conditions are enabled along the 
cell's three spatial directions. 

:Default: ``(False, False, False)``"""
        ...

    @property
    def volume(self) -> float:
        """Read-only property that computes the volume of the three-dimensional simulation cell.
The returned value is equal to the absolute determinant of the 3x3 submatrix formed by the three cell vectors, i.e. the scalar triple product :math:`V=|(\\mathbf{a} \\times \\mathbf{b}) \\cdot \\mathbf{c}|`: 

```python
  assert cell.volume == abs(numpy.linalg.det(cell[0:3,0:3]))
```"""
        ...

    @property
    def volume2D(self) -> float:
        """Read-only property that computes the area of the two-dimensional simulation cell (see :py:attr:`is2D`).
The returned value is equal to the magnitude of the cross-product of the first two cell vectors, i.e. :math:`V_{\\mathrm{2d}} = |\\mathbf{a} \\times \\mathbf{b}|`:

```python
  assert cell.volume2D == numpy.linalg.norm(numpy.cross(cell[:,0], cell[:,1]))
```"""
        ...

    @property
    def vis(self) -> ovito.vis.SimulationCellVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    def __getitem__(self, idx: Any) -> numpy.typing.ArrayLike:
        ...

    def delta_vector(self, ra: numpy.typing.ArrayLike, rb: numpy.typing.ArrayLike, return_pbcvec: bool=False) -> numpy.ndarray:
        """Computes the vector connecting two points :math:`r_a` and :math:`r_b` in a periodic simulation cell by applying the minimum image convention.

The method starts by computing the 3d vector :math:`{\\Delta} = r_b - r_a` for two points :math:`r_a` and :math:`r_b`, which may be located in different images
of the periodic simulation cell. The `minimum image convention <https://en.wikipedia.org/wiki/Periodic_boundary_conditions>`_
is then applied to obtain the new vector :math:`{\\Delta'} = r_b' - r_a`, where the original point :math:`r_b` has been replaced by the periodic image
:math:`r_b'` that is closest to :math:`r_a`, making the vector :math:`{\\Delta'}` as short as possible (in reduced coordinate space).
:math:`r_b'` is obtained by translating :math:`r_b` an integer number of times along each of the three cell directions:
:math:`r_b' = r_b - H*n`, with :math:`H` being the 3x3 cell matrix and :math:`n` being a vector of three integers that are chosen by the
method such that :math:`r_b'` is as close to :math:`r_a` as possible.

Note that the periodic image convention is applied only along those cell directions for which
periodic boundary conditions are enabled (see :py:attr:`pbc` property). For other directions
no shifting is performed, i.e., the corresponding components of :math:`n = (n_x,n_y,n_z)` will always be zero.

The method is able to compute the results for either an individual pair of input points or for two *arrays* of input points. In the latter case,
i.e. if the input parameters *ra* and *rb* are both 2-D arrays of shape *Nx3*, the method returns a 2-D array containing
*N* output vectors. This allows applying the minimum image convention to a large number of point pairs in one function call.

The option *return_pbcvec* lets the method return the vector :math:`n` introduced above as an additional output.
The components of this vector specify the number of times the image point :math:`r_b'` needs to be shifted along each of the three cell directions
in order to bring it onto the original input point :math:`r_b`. In other words, it specifies the number of times the
computed vector :math:`{\\Delta} = r_b - r_a` crosses a periodic boundary of the cell (either in positive or negative direction).
For example, the PBC shift vector :math:`n = (1,0,-2)` would indicate that, in order to get from input point :math:`r_a` to input point :math:`r_b`, one has to cross the
cell boundaries once in the positive x-direction and twice in the negative z-direction. If *return_pbcvec* is True,
the method returns the tuple (:math:`{\\Delta'}`, :math:`n`); otherwise it returns just :math:`{\\Delta'}`.
Note that the vector :math:`n` computed by this method can be used, for instance, to correctly initialize the `Bonds.pbc_vectors`
property for newly created bonds that cross a periodic cell boundary.

:param ra: The Cartesian xyz coordinates of the first input point(s). Either a 1-D array of length 3 or a 2-D array of shape (*N*,3).
:param rb: The Cartesian xyz coordinates of the second input point(s). Must have the same shape as *ra*.
:param bool return_pbcvec: If True, also returns the vector :math:`n`, which specifies how often the vector :math:`(r_b' - r_a)` crosses the periodic cell boundaries.
:returns: The vector :math:`{\\Delta'}` and, optionally, the vector :math:`n`.

Note that there exists also a convenience method `Particles.delta_vector()`,
which should be used in situations where :math:`r_a` and :math:`r_b` are the coordinates of two particles in the simulation cell."""
        ...

class Particles(PropertyContainer):
    """Base: :py:class:`ovito.data.PropertyContainer`

This container object stores the information associated with a system of particles. 
It is typically accessed through the :py:attr:`DataCollection.particles` 
field of a data collection. 
The current number of particles is given by the :py:attr:`~PropertyContainer.count` attribute 
that is inherited from the :py:class:`PropertyContainer` base class. 
The particles may be associated with a set of properties. Each property is represented by a 
:py:class:`Property` data object, that is stored in this property container and is basically an 
array of numeric values of length *N*, where *N* is the number of particles in the system. 
Each property array has a unique name, by which it can be looked up through the dictionary interface of the 
:py:class:`PropertyContainer` base class. 
While the user is free to define arbitrary particle properties, OVITO predefines a set of *standard properties* 
that each have a fixed data layout, meaning and name. They are listed in the table below. 

=================================================== ========== ===================================================
Standard property name                              Data type  Component names
=================================================== ========== ===================================================
:guilabel:`Angular Momentum`                        float      X, Y, Z
:guilabel:`Angular Velocity`                        float      X, Y, Z
:guilabel:`Aspherical Shape`                        float      X, Y, Z
:guilabel:`Centrosymmetry`                          float      
:guilabel:`Charge`                                  float      
:guilabel:`Cluster`                                 int64      
:guilabel:`Color`                                   float      R, G, B
:guilabel:`Coordination`                            int        
:guilabel:`Deformation Gradient`                    float      XX, YX, ZX, XY, YY, ZY, XZ, YZ, ZZ
:guilabel:`Dipole Magnitude`                        float      
:guilabel:`Dipole Orientation`                      float      X, Y, Z
:guilabel:`Displacement Magnitude`                  float      
:guilabel:`Displacement`                            float      X, Y, Z
:guilabel:`DNA Strand`                              int        
:guilabel:`Elastic Deformation Gradient`            float      XX, YX, ZX, XY, YY, ZY, XZ, YZ, ZZ
:guilabel:`Elastic Strain`                          float      XX, YY, ZZ, XY, XZ, YZ
:guilabel:`Force`                                   float      X, Y, Z
:guilabel:`Kinetic Energy`                          float      
:guilabel:`Mass`                                    float      
:guilabel:`Molecule Identifier`                     int64      
:guilabel:`Molecule Type`                           int        
:guilabel:`Nucleobase`                              int        
:guilabel:`Nucleotide Axis`                         float      X, Y, Z
:guilabel:`Nucleotide Normal`                       float      X, Y, Z
:guilabel:`Orientation`                             float      X, Y, Z, W
:guilabel:`Particle Identifier`                     int64      
:guilabel:`Particle Type`                           int        
:guilabel:`Periodic Image`                          int        X, Y, Z
:guilabel:`Position`                                float      X, Y, Z
:guilabel:`Potential Energy`                        float      
:guilabel:`Radius`                                  float      
:guilabel:`Rotation`                                float      X, Y, Z, W
:guilabel:`Selection`                               int        
:guilabel:`Spin`                                    float      
:guilabel:`Strain Tensor`                           float      XX, YY, ZZ, XY, XZ, YZ
:guilabel:`Stress Tensor`                           float      XX, YY, ZZ, XY, XZ, YZ
:guilabel:`Stretch Tensor`                          float      XX, YY, ZZ, XY, XZ, YZ
:guilabel:`Structure Type`                          int        
:guilabel:`Superquadric Roundness`                  float      Phi, Theta
:guilabel:`Torque`                                  float      X, Y, Z
:guilabel:`Total Energy`                            float      
:guilabel:`Transparency`                            float      
:guilabel:`Vector Color`                            float      R, G, B
:guilabel:`Velocity Magnitude`                      float      
:guilabel:`Velocity`                                float      X, Y, Z
=================================================== ========== ==================================================="""

    @property
    def angles(self) -> Optional[PropertyContainer]:
        """A :py:class:`PropertyContainer` storing the list of angles defined for the molecular model (may be ``None``)."""
        ...

    @property
    def bonds(self) -> Optional[PropertyContainer]:
        """The :py:class:`Bonds` data object storing the list of bonds and their properties (may be ``None``)."""
        ...

    @property
    def impropers(self) -> Optional[PropertyContainer]:
        """A :py:class:`PropertyContainer` storing the list of impropers defined for the molecular model (may be ``None``)."""
        ...

    @property
    def dihedrals(self) -> Optional[PropertyContainer]:
        """A :py:class:`PropertyContainer` storing the list of dihedrals defined for the molecular model (may be ``None``)."""
        ...

    @property
    def colors(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Color`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def colors_(self) -> Optional[Property]:
        ...

    @property
    def forces(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Force`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def forces_(self) -> Optional[Property]:
        ...

    @property
    def identifiers(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Particle Identifier`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def identifiers_(self) -> Optional[Property]:
        ...

    @property
    def masses(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Mass`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def masses_(self) -> Optional[Property]:
        ...

    @property
    def particle_types(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Particle Type`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def particle_types_(self) -> Optional[Property]:
        ...

    @property
    def positions(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Position`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def positions_(self) -> Optional[Property]:
        ...

    @property
    def selection(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Selection`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def selection_(self) -> Optional[Property]:
        ...

    @property
    def structure_types(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Structure Type`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def structure_types_(self) -> Optional[Property]:
        ...

    @property
    def velocities(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Velocity`` standard particle property; or ``None`` if that property is undefined."""
        ...

    @property
    def velocities_(self) -> Optional[Property]:
        ...

    @property
    def vis(self) -> ovito.vis.ParticlesVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    def create_particle(self, position: numpy.typing.ArrayLike) -> int:
        """Adds a new particle to the particle system.

:param tuple position: The xyz coordinates for the new particle.
:returns: The index of the newly created particle, i.e. :py:attr:`(Particles.count-1) <ovito.data.PropertyContainer.count>`."""
        ...

    def delta_vector(self, a: numpy.typing.ArrayLike, b: numpy.typing.ArrayLike, cell: SimulationCell, return_pbcvec: bool=False) -> numpy.ndarray:
        """Computes the vector connecting two particles *a* and *b* in a periodic simulation cell by applying the minimum image convention.

This is a convenience wrapper for the `SimulationCell.delta_vector()` method,
which computes the vector between two arbitrary spatial locations :math:`r_a` and :math:`r_b` taking into account periodic
boundary conditions. The version of the method described here takes two particle indices *a* and *b* as input, computing the shortest vector
:math:`{\\Delta} = (r_b - r_a)` between them using the `minimum image convention <https://en.wikipedia.org/wiki/Periodic_boundary_conditions>`_.
Please see the `SimulationCell.delta_vector()` method for further information.

:param a: Zero-based index of the first input particle. This may also be an array of particle indices.
:param b: Zero-based index of the second input particle. This may also be an array of particle indices with the same length as *a*.
:param cell: The :py:class:`SimulationCell` object that defines the periodic domain. Typically, `DataCollection.cell` is used here.
:param bool return_pbcvec: If True, also returns the vector :math:`n`, which specifies how often the computed particle-to-particle vector crosses the cell's face.
:returns: The delta vector and, optionally, the vector :math:`n`."""
        ...

class Bonds(PropertyContainer):
    """Base: :py:class:`ovito.data.PropertyContainer`

Stores the list of bonds and their properties. A :py:class:`Bonds` object is always part of a parent :py:class:`Particles` object. 
You can access it as follows: 

```python
  data = pipeline.compute()
  print("Number of bonds:", data.particles.bonds.count)
```

The :py:class:`Bonds` class inherits the :py:attr:`~PropertyContainer.count` attribute from its :py:class:`PropertyContainer` base class. This attribute returns the number of bonds. 

Bond properties

Bonds can be associated with arbitrary *bond properties*, which are managed in the :py:class:`Bonds` container 
as a set of :py:class:`Property` data arrays. Each bond property has a unique name by which it can be looked up: 

```python
  print("Bond property names:")
  print(data.particles.bonds.keys())
  if 'Length' in data.particles.bonds:
      length_prop = data.particles.bonds['Length']
      assert(len(length_prop) == data.particles.bonds.count)
```

New bond properties can be added using the :py:meth:`PropertyContainer.create_property` method. 

Bond topology

The ``Topology`` bond property, which is always present, 
defines the connectivity between particles in the form of a *N* x 2 array of indices into the :py:class:`Particles` array. 
In other words, each bond is defined by a pair of particle indices. 

```python
  for a,b in data.particles.bonds.topology:
      print("Bond from particle %i to particle %i" % (a,b))
```

Note that the bonds of a system are not stored in any particular order. 
If you need to enumerate all bonds connected to a certain particle, you can use the :py:class:`BondsEnumerator` utility class for that. 

Bonds visualization

The :py:class:`Bonds` data object has a :py:class:`BondsVis` element attached to it, 
which controls the visual appearance of the bonds in rendered images. It can be accessed through the :py:attr:`~DataObject.vis` 
attribute: 

```python
  data.particles.bonds.vis.enabled = True
  data.particles.bonds.vis.flat_shading = True
  data.particles.bonds.vis.width = 0.3
```

Computing bond vectors

Since each bond is defined by two indices into the particles array, we can use these indices to determine the corresponding spatial 
bond *vectors* connecting the particles. They can be computed from the positions of the particles: 

```python
  topology = data.particles.bonds.topology
  positions = data.particles.positions
  bond_vectors = positions[topology[:,1]] - positions[topology[:,0]]
```

Here, the first and the second column of the bonds topology array are used to index into the particle positions array. 
The subtraction of the two indexed arrays yields the list of bond vectors. Each vector in this list points 
from the first particle to the second particle of the corresponding bond. 

Finally, we may have to correct for the effect of periodic boundary conditions when a bond 
connects two particles on opposite sides of the box. OVITO keeps track of such cases by means of the 
the special ``Periodic Image`` bond property. It stores a shift vector for each bond, specifying the directions in which the bond 
crosses periodic boundaries. We make use of this information to correct the bond vectors computed above.
This is done by adding the product of the cell matrix and the shift vectors from the ``Periodic Image`` bond property:

```python
  bond_vectors += numpy.dot(data.cell[:3,:3], data.particles.bonds.pbc_vectors.T).T
```

The shift vectors array is transposed here to facilitate the transformation 
of the entire array of vectors with a single 3x3 cell matrix. 
To summarize: In the two code snippets above we have performed 
the following calculation of the *unwrapped* vector :math:`\\mathbf{v}` for every bond (*a*, *b*) in parallel: 

   :math:`\\mathbf{v} = \\mathbf{x}_b - \\mathbf{x}_a + \\mathbf{H} \\cdot (n_x, n_y, n_z)^{T}`,

with :math:`\\mathbf{H}` denoting the simulation cell matrix and :math:`(n_x, n_y, n_z)` the bond's PBC shift vector.

Standard bond properties

The following standard properties are defined for bonds:

================================ ======================== ========== =======================
Property name                    Python attribute name    Data type  Component names
================================ ======================== ========== =======================
:guilabel:`Bond Type`            :py:attr:`bond_types`    int        
:guilabel:`Color`                :py:attr:`colors`        float      R, G, B
:guilabel:`Length`                                        float       
:guilabel:`Particle Identifiers`                          int64      A, B\\
:guilabel:`Periodic Image`       :py:attr:`pbc_vectors`   int        X, Y, Z 
:guilabel:`Selection`            :py:attr:`selection`     int         
:guilabel:`Topology`             :py:attr:`topology`      int64      1, 2\\
:guilabel:`Transparency`                                  float       
:guilabel:`Width`                                         float       
================================ ======================== ========== ======================="""

    @property
    def bond_types(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Bond Type`` standard bond property; or ``None`` if that property is undefined."""
        ...

    @property
    def bond_types_(self) -> Optional[Property]:
        ...

    @property
    def colors(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Color`` standard bond property; or ``None`` if that property is undefined."""
        ...

    @property
    def colors_(self) -> Optional[Property]:
        ...

    @property
    def pbc_vectors(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Periodic Image`` standard bond property; or ``None`` if that property is undefined."""
        ...

    @property
    def pbc_vectors_(self) -> Optional[Property]:
        ...

    @property
    def selection(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Selection`` standard bond property; or ``None`` if that property is undefined."""
        ...

    @property
    def selection_(self) -> Optional[Property]:
        ...

    @property
    def topology(self) -> Optional[Property]:
        """The :py:class:`Property` data array for the ``Topology`` standard bond property; or ``None`` if that property is undefined."""
        ...

    @property
    def topology_(self) -> Optional[Property]:
        ...

    @property
    def vis(self) -> ovito.vis.BondsVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    def create_bond(self, a: int, b: int, type: Optional[int]=None, pbcvec: Optional[numpy.typing.ArrayLike]=None) -> int:
        """Creates a new bond between two particles *a* and *b*, both parameters being indices into the :py:class:`Particles` list
to which this :py:class:`Bonds` container belongs.

:param int a: Index of first particle connected by the new bond. Particle indices start at 0.
:param int b: Index of second particle connected by the new bond.
:param int type: Optional type ID to be assigned to the new bond. This value will be stored to the :py:attr:`bond_types` array.
:param tuple pbcvec: Three integers specifying the bond's crossings of periodic cell boundaries. The information will be stored in the :py:attr:`pbc_vectors` array.
:returns: The index of the newly created bond, i.e. :py:attr:`(Bonds.count-1) <ovito.data.PropertyContainer.count>`.

The method does *not* check if there already is an existing bond connecting the same pair of particles.

The method does *not* check if the particle indices *a* and *b* do exist. Thus, it is your responsibility to ensure that both indices
are in the range 0 to :py:attr:`(Particles.count-1) <ovito.data.PropertyContainer.count>`.

In case the :py:class:`SimulationCell` has periodic boundary conditions enabled, and the two particles connected by the bond are located in different periodic images,
make sure you provide the *pbcvec* argument. It is required so that OVITO does not draw the bond as a direct line from particle *a* to particle *b* but as a line passing through
the periodic cell faces. You can use the `Particles.delta_vector()` function to compute
*pbcvec* or use the ``pbc_shift`` vector returned by the :py:class:`CutoffNeighborFinder` utility."""
        ...

class SurfaceMesh(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

A triangle mesh representation of a surface or, more precisely, a two-dimensional manifold that is closed and orientable. 
Typically, surface meshes are produced within OVITO by modifiers such as the :py:class:`ConstructSurfaceModifier`, 
:py:class:`CreateIsosurfaceModifier` or :py:class:`CoordinationPolyhedraModifier`. 
Please see the user manual page on surface meshes for a more in-depth description
of this geometric data structure.

Periodic domains

Surface meshes may be embedded in periodic domains, i.e. in simulation cells with periodic boundary conditions applied. 
That means triangles of a surface mesh can be connected to vertices on opposite sides of a simulation box and wrap around correctly. 
OVITO takes care of computing the intersections of the periodic surface with the box boundaries and automatically produces a non-periodic representation 
of the mesh when it comes to visualization, which is cut off at the box boundaries. 

The spatial domain the surface mesh is embedded in is specified as a :py:class:`SimulationCell` object attached to the 
:py:class:`SurfaceMesh`. It is stored in the :py:attr:`SurfaceMesh.domain` field. Note that this :py:class:`SimulationCell` 
may, in some situations, not be identical to the global simulation :py:attr:`cell` set for the :py:class:`DataCollection`.

Spatial regions

Being a closed, orientable manifold a surface mesh divides space into two or more spatial regions. 
For example, if the surface mesh was constructed by the :py:class:`ConstructSurfaceModifier` from a set of input particles, 
then the space enclosed by the surface is the "filled" region and the exterior space is the "empty" region containing no particles. 

In general, the :py:class:`SurfaceMesh` class manages a variable list of :py:attr:`regions`, each being identified by a numeric, zero-based index.
The :py:meth:`locate_point` method allows to determine which spatial region some point in space belongs to.

It may be that a surface mesh is degenerate. In such a case there is only one spatial region filling entire space. For example, 
when there are no input particles, the :py:class:`ConstructSurfaceModifier` cannot construct a regular surface mesh and
the "empty" region fills the entire simulation cell. Conversely, if the periodic simulation cell is completely filled with particles,
the "filled" region covers the entire simulation domain and the generated surface mesh will consist of no vertices or faces, i.e., it is degenerate.
To discriminate between the two situations, the :py:class:`SurfaceMesh` class has a :py:attr:`space_filling_region` field, which specifies
the spatial region that fills entire space in cases where the mesh is degenerate.

File export

A surface mesh can be exported to a geometry file in the form of a triangle mesh using OVITO's :py:func:`export_file` function. 
To this end, a non-periodic version is produced by truncating triangles at the domain boundaries and generating "cap polygons" filling the holes that 
occur at the intersection of the surface with periodic domain boundaries. The following example code writes a VTK geometry file 
(``vtk/trimesh`` export format):

```python
  from ovito.io import import_file, export_file
  from ovito.data import SurfaceMesh
  from ovito.modifiers import ConstructSurfaceModifier
  
  # Load a particle set and construct the surface mesh:
  pipeline = import_file("input/simulation.dump")
  pipeline.modifiers.append(ConstructSurfaceModifier(radius = 2.8))
  mesh = pipeline.compute().surfaces['surface']
  
  # Export the mesh to a VTK file for visualization with ParaView.
  export_file(mesh, 'output/surface_mesh.vtk', 'vtk/trimesh')
```

Cutting planes

A set of *cutting planes* can be assigned to a :py:class:`SurfaceMesh` to cut away parts of the mesh for visualization purposes.
This may be useful to e.g. cut a hole into a closed surface allowing to look inside the enclosed volume.
The :py:class:`SurfaceMesh` objects manages a list of cutting planes, which are accessible through the :py:meth:`.get_cutting_planes` and :py:meth:`.set_cutting_planes`
methods. Note that the cuts are non-destructive and get performed only on the transient, non-periodic version of the mesh generated during image rendering or when exporting the mesh to a file.
The original data structures of the mesh are not affected.
The :py:class:`SliceModifier`, which can act on a :py:class:`SurfaceMesh`, performs the slice by simply adding a new entry to the :py:class:`SurfaceMesh`'s
list of cutting planes.

Mesh data access

The methods :py:meth:`.get_vertices`, :py:meth:`.get_faces` and :py:meth:`.get_face_adjacency` methods provide access to the internal data of the
surface mesh."""

    @property
    def domain(self) -> Optional[SimulationCell]:
        """The :py:class:`SimulationCell` describing the (possibly periodic) domain which this surface mesh is embedded in. Note that this cell generally is independent of and may be different from the :py:attr:`cell` found in the :py:class:`DataCollection`."""
        ...

    @property
    def faces(self) -> PropertyContainer:
        """The :py:class:`PropertyContainer` storing the properties of the faces of the mesh."""
        ...

    @property
    def vertices(self) -> PropertyContainer:
        """The :py:class:`PropertyContainer` storing the vertex properties of the mesh, including the vertex coordinates."""
        ...

    @property
    def regions(self) -> PropertyContainer:
        """The :py:class:`PropertyContainer` storing the properties of the spatial regions of the mesh."""
        ...

    @property
    def space_filling_region(self) -> int:
        """Indicates the index of the spatial region that fills the entire domain in case the surface is degenerate, i.e. the mesh has zero faces. The invalid index -1 is typically associated with the empty (exterior) region."""
        ...

    @property
    def vis(self) -> ovito.vis.SurfaceMeshVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    def get_cutting_planes(self) -> numpy.typing.NDArray:
        """Returns a *N* x 4 array containing the definitions of the *N* cutting planes attached to this :py:class:`SurfaceMesh`. 

Each plane is defined by its unit normal vector and a signed displacement magnitude, which determines the plane's distance from the coordinate origin along the normal, giving four numbers per plane in total. Those parts of the surface mesh which are on the positive side of the plane (in the direction the normal vector) are cut away. 

Note that the returned Numpy array is a copy of the internal data stored by the :py:class:`SurfaceMesh`."""
        ...

    def get_face_adjacency(self) -> numpy.typing.NDArray:
        """Returns a *M* x 3 array listing the indices of the three faces that are adjacent to each of the *M* triangle faces in the mesh. This information can be used to traverse the neighbors of triangle faces. Every triangle face has exactly three neighbors, because surface meshes are closed manifolds."""
        ...

    def get_faces(self) -> numpy.typing.NDArray:
        """Returns a *M* x 3 array with the vertex indices of the *M* triangles in the mesh. Note that the returned Numpy array is a copy of the internal data stored by the :py:class:`SurfaceMesh`. Also keep in mind that triangle faces can cross the domain boundaries if the periodic boundary conditions are used."""
        ...

    def get_vertices(self) -> numpy.typing.NDArray:
        """Returns a *N* x 3 array with the xyz coordinates of the *N* vertices in the mesh. Note that the returned Numpy array is a copy of the internal data stored by the :py:class:`SurfaceMesh`."""
        ...

    def locate_point(self, pos: ovito.vis.Vector3, eps: float=1e-06) -> Optional[int]:
        """Determines the index of the spatial region that contains the given location in 3-D space. Note that region index -1 is typically reserved for the empty/exterior region, which doesn't contain any atoms or particles. Regions starting at index 0 are used for filled or interior regions. 

The parameter *eps* is used as a precision threshold to detect cases where the query point is positioned exactly on the surface itself, i.e. on the boundary between two spatial regions. Such a condition is indicated by the special return value ``None``. You can set *eps* to 0.0 to disable the point-on-boundary test. Then the method will never yield ``None`` as a result. 

:param pos: The (x,y,z) coordinates of the query point
:param eps: Numerical precision threshold for point-on-boundary test
:return: The ID of the spatial region containing *pos*; or ``None`` if *pos* is exactly on the dividing surface between two regions"""
        ...

class DislocationNetwork(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

This data object stores the network of dislocation lines extracted by a :py:class:`DislocationAnalysisModifier`. You can access it through the :py:attr:`DataCollection.dislocations` field. 

The dislocation network is associated with a :py:class:`DislocationVis` element controlling the visual appearance of the dislocation lines. It can be accessed through the :py:attr:`~DataObject.vis` attribute of the :py:class:`~DataObject` base class.

Example:

```python
  from ovito.io import import_file, export_file
  from ovito.modifiers import DislocationAnalysisModifier
  from ovito.data import DislocationNetwork
  
  import ovito
  ovito.enable_logging()
  
  pipeline = import_file("input/simulation.dump")
  
  # Extract dislocation lines from a crystal with diamond structure:
  modifier = DislocationAnalysisModifier()
  modifier.input_crystal_structure = DislocationAnalysisModifier.Lattice.CubicDiamond
  pipeline.modifiers.append(modifier)
  data = pipeline.compute()
  
  total_line_length = data.attributes['DislocationAnalysis.total_line_length']
  cell_volume = data.attributes['DislocationAnalysis.cell_volume']
  print("Dislocation density: %f" % (total_line_length / cell_volume))
  
  # Print list of dislocation lines:
  print("Found %i dislocation segments" % len(data.dislocations.segments))
  for segment in data.dislocations.segments:
      print("Segment %i: length=%f, Burgers vector=%s" % (segment.id, segment.length, segment.true_burgers_vector))
      print(segment.points)
  
  # Export dislocation lines to a CA file:
  export_file(pipeline, "output/dislocations.ca", "ca")
  
  # Or export dislocations to a ParaView VTK file:
  export_file(pipeline, "output/dislocations.vtk", "vtk/disloc")
```

File export

A dislocation network can be written to a data file in the form of polylines using the :py:func:`ovito.io.export_file` function (select the ``vtk/disloc`` output format). During export, a non-periodic version is produced by clipping dislocation lines at the domain boundaries."""

    @property
    def vis(self) -> ovito.vis.DislocationVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

    @property
    def segments(self) -> Sequence[DislocationSegment]:
        """The list of dislocation segments in this dislocation network. This list-like object is read-only and contains :py:class:`DislocationSegment` objects."""
        ...

    def set_segment(self, index: int, true_burgers_vector: Optional[ovito.vis.Vector3]=None, cluster_id: Optional[int]=None, points: Optional[numpy.typing.ArrayLike]=None, custom_color: Optional[ovito.vis.Color]=None) -> None:
        """This method allows you to change the data fields of individual dislocation lines. Fields for which no new value is specified will keep their current values. 

:param index: The zero-based index of the dislocation line in the :py:attr:`segments` array to be modified.
:param true_burgers_vector: The lattice-space Burgers vector (:py:attr:`true_burgers_vector`) to be assigned to the dislocation line.
:param cluster_id: The numeric ID of the crystallite cluster the dislocation line is embedded in.
:param points: A *N* x 3 NumPy array with the Cartesian coordinates of the dislocation line vertices.
:param custom_color: RGB color to be used for rendering the line instead of the automatically determined color.


Example of a :ref:`user-defined modifier function <writing_custom_modifiers>` manipulating the dislocation line data:

```python
  import numpy as np
  
  def modify(frame, data):
  
      # Flip Burgers vector and line sense of each dislocation:
      for index, seg in enumerate(data.dislocations.segments):
          data.dislocations_.set_segment(index, 
              true_burgers_vector = np.negative(seg.true_burgers_vector), 
              points = np.flipud(seg.points))
  
      # Highlight all 1/6[121] dislocations in a red color:
      for index, seg in enumerate(data.dislocations.segments):
          if np.allclose(seg.true_burgers_vector, (1/6, 2/6, 1/6)):
              data.dislocations_.set_segment(index, custom_color=(1, 0, 0))
```"""
        ...

class TriangleMesh(DataObject):
    """Base: :py:class:`ovito.data.DataObject`

This data object type stores a triangle mesh describing general polyhedral geometry, which may be visualized side by side with the simulation data. 
Typically, triangle meshes are imported from an external geometry file using the :py:func:`import_file` function.
See also the corresponding page of the user manual for more information on triangle meshes. 
This class has no parameters or methods. 

See also the :py:class:`SurfaceMesh` class, which is another data object type in OVITO representing surface geometries.
In contrast to triangle meshes, however, surface meshes can live inside of periodic simulation domains and are true orientable, closed manifolds.
Furthermore, surface meshes can store arbitrary per-vertex and per-face property data -- something triangle meshes cannot do. 

The visual appearance of the triangle mesh in rendered images is controlled through the :py:class:`TriangleMeshVis`
element attached to this data object (see base class :py:class:`DataObject`)."""

    @property
    def vis(self) -> ovito.vis.TriangleMeshVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

class TrajectoryLines(PropertyContainer):
    """Base: :py:class:`ovito.data.PropertyContainer`

Data object that stores the trajectory lines of a set of particles, 
which have been traced by the :py:class:`GenerateTrajectoryLinesModifier`. 
It is typically part of a pipeline's output data collection, 
from where it can be accessed via the `DataCollection.trajectories` field. 

A :py:class:`TrajectoryLines` object has an associated :py:class:`TrajectoryVis`
element, which controls the visual appearance of the trajectory lines in rendered images.
This visual element is accessible through the :py:attr:`~DataObject.vis` attribute of the base class."""

    @property
    def positions(self) -> Property:
        """The :py:class:`Property` data array storing the XYZ coordinates of the line vertices."""
        ...

    @property
    def positions_(self) -> Property:
        ...

    @property
    def particle_ids(self) -> Optional[Property]:
        """The :py:class:`Property` data array storing the particle IDs of the line vertices."""
        ...

    @property
    def particle_ids_(self) -> Optional[Property]:
        ...

    @property
    def time_stamps(self) -> Property:
        """The :py:class:`Property` data array storing the time stamps of the line vertices."""
        ...

    @property
    def time_stamps_(self) -> Property:
        ...

    @property
    def vis(self) -> ovito.vis.TrajectoryVis:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...

class CutoffNeighborFinder:
    """A utility class that computes particle neighbor lists.

This class lets you iterate over all neighbors of a particle that are located within a specified spherical cutoff.
You can use it to build neighbors lists or perform computations that require neighbor vector information.

The constructor takes a positive cutoff radius and a `DataCollection` 
providing the input particles and the :py:class:`SimulationCell` (needed for periodic systems).

Once the :py:class:`CutoffNeighborFinder` has been constructed, you can call its :py:meth:`.find` method to 
iterate over the neighbors of a particle, for example:

```python
  from ovito.io import import_file
  from ovito.data import CutoffNeighborFinder
  
  # Load input simulation file.
  pipeline = import_file("input/simulation.dump")
  data = pipeline.compute()
  
  # Initialize neighbor finder object:
  cutoff = 3.5
  finder = CutoffNeighborFinder(cutoff, data)
  
  # Prefetch the property array containing the particle type information:
  ptypes = data.particles.particle_types
  
  # Loop over all particles:
  for index in range(data.particles.count):
      print("Neighbors of particle %i:" % index)
  
      # Iterate over the neighbors of the current particle:
      for neigh in finder.find(index):
          print(neigh.index, neigh.distance, neigh.delta, neigh.pbc_shift)
  
          # The index can be used to access properties of the current neighbor, e.g.
          type_of_neighbor = ptypes[neigh.index]
```

Note: In case you rather want to determine the *N* nearest neighbors of a particle,
use the :py:class:`NearestNeighborFinder` class instead."""

    class Neighbor:
        index: int
        distance: float
        distance_squared: float
        delta: Tuple[float, float, float]
        pbc_shift: Tuple[int, int, int]

    def __init__(self, cutoff: float, data_collection: DataCollection) -> None:
        """This is the constructor."""
        ...

    def find(self, index: int) -> Iterator[Neighbor]:
        """Returns an iterator over all neighbors of the given particle.
 
:param int index: The zero-based index of the central particle whose neighbors should be enumerated.
:returns: A Python iterator that visits all neighbors of the central particle within the cutoff distance. 
          For each neighbor the iterator returns an object with the following property fields:
          
              * index: The zero-based global index of the current neighbor particle.
              * distance: The distance of the current neighbor from the central particle.
              * distance_squared: The squared neighbor distance.
              * delta: The three-dimensional vector connecting the central particle with the current neighbor (taking into account periodicity).
              * pbc_shift: The periodic shift vector, which specifies how often each periodic boundary of the simulation cell is crossed when going from the central particle to the current neighbor.

The `index` value returned by the iterator can be used to look up properties of the neighbor particle as demonstrated in the example above.

Note that all periodic images of particles within the cutoff radius are visited. Thus, the same particle index may appear multiple times in the neighbor
list of the central particle. In fact, the central particle may be among its own neighbors in a small periodic simulation cell.
However, the computed vector (``delta``) and PBC shift (``pbc_shift``) will be unique for each visited image of the neighbor particle."""
        ...

    def find_at(self, coords: ovito.vis.Vector3) -> Iterator[Neighbor]:
        """Returns an iterator over all particles located within the spherical range of the given center position. In contrast to :py:meth:`find` this method can search for neighbors around arbitrary 
spatial locations, which don't have to coincide with any physical particle position.
 
:param coords: A (x,y,z) coordinate triplet specifying the center location around which to search for particles.
:returns: A Python iterator enumerating all particles within the cutoff distance. 
          For each neighbor the iterator returns an object with the following properties:
          
              * index: The zero-based global index of the current neighbor particle.
              * distance: The distance of the current particle from the center position.
              * distance_squared: The squared distance.
              * delta: The three-dimensional vector from the center to the current neighbor (taking into account periodicity).
              * pbc_shift: The periodic shift vector, which specifies how often each periodic boundary of the simulation cell is crossed when going from the center point to the current neighbor.

The index value returned by the iterator can be used to look up properties of the neighbor particle as demonstrated in the example above.

Note that all periodic images of particles within the cutoff radius are visited. Thus, the same particle index may appear multiple times in the neighbor list. 
However, the computed vector (``delta``) and image offset (``pbc_shift``) will be unique for each visited image of a neighbor particle."""
        ...

    def neighbor_distances(self, index: int) -> numpy.typing.NDArray[numpy.float64]:
        """Returns the list of distances between some central particle and all its neighbors within the cutoff range. 

:param int index: The 0-based index of the central particle whose neighbors should be enumerated.
:returns: NumPy array containing the radial distances to all neighbor particles within the cutoff range (in arbitrary order).

This method is equivalent to the following code, but performance is typically better::

   def neighbor_distances(index):
       distances = []
       for neigh in finder.find(index):
           distances.append(neigh.distance)
       return numpy.asarray(distances)"""
        ...

    def neighbor_vectors(self, index: int) -> numpy.typing.NDArray[numpy.float64]:
        """Returns the list of vectors from some central particle to all its neighbors within the cutoff range. 

:param int index: The 0-based index of the central particle whose neighbors should be enumerated.
:returns: Two-dimensional NumPy array containing the vectors to all neighbor particles within the cutoff range (in arbitrary order).


The method is equivalent to the following code, but performance is typically better::

   def neighbor_vectors(index):
       vecs = []
       for neigh in finder.find(index):
           vecs.append(neigh.delta)
       return numpy.asarray(vecs)"""
        ...

class NearestNeighborFinder:
    """A utility class that finds the *N* nearest neighbors of a particle. 

.. seealso::

  To find all neighbors within a spherical cutoff region around another particle, 
  use :py:class:`CutoffNeighborFinder` instead.

The constructor takes the requested number of nearest neighbors, *N*, and a `DataCollection`
containing the input particles and the optional simulation cell.
*N* must be a positive integer not greater than 30, which is the built-in maximum of neighbors supported by this class.

.. note::

  Keep in mind that, if the system contains only *N* particles or less, and if the simulation does not use periodic boundary conditions, then the 
  neighbor finder will return less than the requested number of nearest neighbors.

Once the :py:class:`NearestNeighborFinder` has been initialized, you can call its :py:meth:`.find` method to
iterate over the sorted list of nearest neighbors of a given central particle:

```python
  # Set up a neighbor finder for visiting the 12 closest neighbors of each particle.
  finder = NearestNeighborFinder(12, data)
  
  # Loop over all input particles:
  for index in range(data.particles.count):
      print("Nearest neighbors of particle %i:" % index)
      # Iterate over the neighbors of the current particle, starting with the closest:
      for neigh in finder.find(index):
          print(neigh.index, neigh.distance, neigh.delta)
          # The index can be used to access properties of the current neighbor, e.g.
          type_of_neighbor = data.particles.particle_types[neigh.index]
```

In addition, the class provides the :py:meth:`find_at` method, which determines the *N* nearest particles around some
arbitrary spatial location:

```python
  # Visit particles closest to some spatial point (x,y,z):
  xyz_coords = (0.0, 0.0, 0.0)
  for neigh in finder.find_at(xyz_coords):
      print(neigh.index, neigh.distance, neigh.delta)
```"""

    class Neighbor:
        index: int
        distance: float
        distance_squared: float
        delta: Tuple[float, float, float]

    def __init__(self, N: int, data_collection: DataCollection) -> None:
        """Initializes the neighbor finder facility."""
        ...

    def find(self, index: int) -> Iterator[Neighbor]:
        """Returns an iterator that visits the *N* nearest neighbors of the given particle in order of ascending distance.

:param int index: The zero-based index of the central particle whose neighbors should be determined.
:returns: A Python iterator that visits the *N* nearest neighbors of the central particle in order of ascending distance.
          For each neighbor being visited, the iterator returns an object having the following attributes:

              * index: The global index of the current neighbor particle.
              * distance: The distance of the current neighbor from the central particle.
              * distance_squared: The squared neighbor distance.
              * delta: The three-dimensional vector connecting the central particle with the current neighbor (correctly taking into account periodic boundary conditions).

The index can be used to look up properties of the neighbor particle as demonstrated in the first example code above.

Note that several periodic images of the same particle may be visited if the periodic simulation cell is sufficiently small.
Then the same particle index will appear more than once in the neighbor list. In fact, the central particle may be among its own neighbors in a sufficiently small periodic simulation cell.
However, the computed neighbor vector (`delta`) will be unique for each image of a neighboring particle.

The number of neighbors actually visited may be smaller than the requested number, *N*, if the
system contains too few particles and is non-periodic.

Note that the :py:meth:`!find()` method will not find other particles located exactly at the same spatial position as the central particle for technical reasons.
To find such particles too, which are positioned exactly on top of each other, use :py:meth:`.find_at` instead."""
        ...

    def find_at(self, coords: ovito.vis.Vector3) -> Iterator[Neighbor]:
        """Returns an iterator that visits the *N* nearest particles around a spatial point given by *coords* in order of ascending distance.
Unlike the :py:meth:`find` method, which queries the nearest neighbors of a physical particle, :py:meth:`find_at` allows
searching for nearby particles at arbitrary locations in space.

:param coords: A coordinate triplet (x,y,z) specifying the spatial location where the *N* nearest particles should be queried.
:returns: A Python iterator that visits the *N* nearest neighbors in order of ascending distance.
          For each visited particle the iterator returns an object with the following attributes:

              * index: The index of the current particle (starting at 0).
              * distance: The distance of the current neighbor from the query location.
              * distance_squared: The squared distance to the query location.
              * delta: The three-dimensional vector from the query point to the current particle (correctly taking into account periodic boundary conditions).

If there is a particle located exactly at the query location *coords*, it will be among the returned neighbors.
This is in contrast to the :py:meth:`find` function, which skips the central particle itself.

The number of neighbors actually visited may be smaller than the requested number, *N*, if the
system contains too few particles and is non-periodic."""
        ...

    def find_all(self, indices: Optional[Sequence[int]]=None) -> Tuple[numpy.typing.NDArray[numpy.int64], numpy.typing.NDArray[numpy.float64]]:
        """Finds the *N* nearest neighbors of each particle in the system or of the subset of particles specified by *indices*. 
This is the batch-processing version of :py:meth:`find`, allowing you to efficiently compute the neighbor lists and neighbor vectors of several 
particles at once, without explicit for-loop and by making use of all parallel processor cores. 

The method returns two NumPy arrays:

``neigh_idx`` : NumPy array of shape (*M*, *N*) storing the indices of neighbor particles, 
with *M* equal to *len(indices)* or, if *indices* is *None*, the total number of particles in the system. 
*N* refers to the number of nearest neighbors requested in the :py:class:`NearestNeighborFinder` constructor. 
The computed indices in this array can be used to look up properties of neighbor particles in the global :py:class:`Particles` object.

``neigh_vec`` : NumPy array of shape (*M*, *N*, 3) storing the xyz components of the three-dimensional neighbor vectors ("delta"), 
which connect the *M* central particles with their *N* respective nearest neighbors. 

:param indices: List of zero-based particle indices for which the neighbor lists should be computed. 
                If left unspecified, neighbor lists will be computed for every particle in the system.
:returns: ``(neigh_idx, neigh_vec)``

.. tip::

   To compute all pair-wise distances in one go, i.e. the 2-norms of the neighbor vectors, you can do::

      distances = numpy.linalg.norm(neigh_vec, axis=2)   # Yields (M,N) array of neighbor distances"""
        ...

class BondsEnumerator:
    """Utility class that permits efficient iteration over the bonds connected to specific particles. 

The constructor takes a :py:class:`Bonds` object as input. 
From the generally unordered list of bonds, the :py:class:`BondsEnumerator` will build a lookup table for quick enumeration  
of bonds of particular particles. 

All bonds connected to a specific particle can be subsequently visited using the :py:meth:`.bonds_of_particle` method. 

Warning: Do not modify the underlying :py:class:`Bonds` object while the :py:class:`BondsEnumerator` is in use. 
Adding or deleting bonds would render the internal lookup table of the :py:class:`BondsEnumerator` invalid. 

Usage example

.. literalinclude:: ../example_snippets/bonds_enumerator.py"""

    def __init__(self, bonds: Bonds) -> None:
        """Initializes the bond enumerator.

:param Bonds bonds: The data object holding the (unsorted) list of input bonds."""
        ...

    def bonds_of_particle(self, index: int) -> Iterator[int]:
        """Returns an iterator yielding the indices of the bonds connected to the given particle. The indices can be used to index into the :py:class:`Property` arrays of the :py:class:`Bonds` object."""
        ...

class DataTable(PropertyContainer):
    """Base: :py:class:`ovito.data.PropertyContainer`

This data object type represents a series of data points and is used for generating histogram plots and other 2d data graphs. 
A data table consists of an array of y-values and, optionally, an array of corresponding x-values, one for each data point. 
Both arrays are standard :py:class:`Property` objects managed by the data table, which is a :py:class:`PropertyContainer` 
subclass.

If the :py:attr:`x` data array is not present, the x-coordinates of the data points are implicitly determined by the table's
:py:attr:`interval`, which specifies a value range along the x-axis over which the data points are evenly distributed. 
This is used, for example, for histograms having equisized bins. 

Data tables generated by modifiers such as :py:class:`CoordinationAnalysisModifier` and :py:class:`HistogramModifier`
are accessible through the `DataCollection.tables` dictionary of the data collection.
Please see the documentation of a modifier to find out what data table(s) it produces.

Examples:

The following code examples demonstrate how to create a new :py:class:`DataTable` and
fill it with data values in Python. You can use these techniques, for instance, to write custom modifier functions 
that output their results as data plots in OVITO.

To create a simple x-y scatter point plot: 

```python
  # Create a DataTable object, setting the title and plot type:
  table = DataTable(title='My Scatter Plot', plot_mode=DataTable.PlotMode.Scatter)
  # Set the x- and y-coordinates of the data points:
  table.x = table.create_property('X coordinates', data=numpy.linspace(0.0, 10.0, 50))
  table.y = table.create_property('Y coordinates', data=numpy.cos(table.x))
  # Add the DataTable to the output DataCollection:
  data.objects.append(table)
```

Note how the :py:meth:`create_property` method is used here to create two :py:class:`Property`
objects holding the x- and y-coordinates of the data points. These property objects are then set as :py:attr:`x` and :py:attr:`y` arrays
of the :py:class:`DataTable`.

Generating a graph with several line plots requires creating a *vector* property object for the :py:attr:`y` array of the :py:class:`DataTable`:

```python
  table = DataTable(title='Trigonometric functions', plot_mode=DataTable.PlotMode.Line)
  table.x = table.create_property('Parameter x', data=numpy.linspace(0.0, 14.0, 100))
  # Use the x-coords to compute two y-coords per data point: y(x) = (cos(x), sin(x)) 
  y1y2 = numpy.stack((numpy.cos(table.x), numpy.sin(table.x)), axis=1)
  table.y = table.create_property('f(x)', data=y1y2, components=['cos(x)', 'sin(x)'])
  data.objects.append(table)
```

To generate a bar chart, the :py:attr:`x` property must be filled with numeric IDs 0,1,2,3,... denoting the individual bars. 
Each bar is then assigned a text label by creating a :py:class:`ElementType` instance and adding it to the `Property.types` list:

```python
  table = DataTable(title='My Bar Chart', plot_mode=DataTable.PlotMode.BarChart)
  table.x = table.create_property('Structure Type', data=[0, 1, 2, 3])
  table.x.types.append(ElementType(id=0, name='Other'))
  table.x.types.append(ElementType(id=1, name='FCC'))
  table.x.types.append(ElementType(id=2, name='HCP'))
  table.x.types.append(ElementType(id=3, name='BCC'))
  table.y = table.create_property('Count', data=[65, 97, 10, 75])
  data.objects.append(table)
```

For histogram plots, one can specify the complete range of values covered by the histogram by setting the table's :py:attr:`interval` property. 
The bin counts must be assigned to the table's :py:attr:`y` property. The number of elements in the :py:attr:`y` property array, together with the
:py:attr:`interval`, determine the number of histogram bins and their width:

```python
  table = DataTable(title='My Histogram', plot_mode=DataTable.PlotMode.Histogram)
  table.y = table.create_property('Counts', data=[65, 97, 10, 75])
  table.interval = (0.0, 2.0)   # Four histogram bins of width 0.5 each.
  table.axis_label_x = 'Values' # Set the x-axis label of the plot.
  data.objects.append(table)
```

If you are going to refer to the data table after it has been inserted into the `DataCollection.objects`
list, you should give it a unique :py:attr:`identifier` at construction time, as shown in the following example:

```python
  def modify(frame, data):
      table = DataTable(identifier='trig-func', title='My Plot', plot_mode=DataTable.PlotMode.Line)
      table.x = table.create_property('X coords', data=numpy.linspace(0.0, 10.0, 50))
      table.y = table.create_property('Y coords', data=numpy.cos(frame * table.x))
      data.objects.append(table)
  
  pipeline.modifiers.append(modify)
  export_file(pipeline, 'output/data.*.txt', 'txt/table', key='trig-func', multiple_frames=True)
```

Here, the :ref:`user-defined modifier function <writing_custom_modifiers>` outputs a table to the data pipeline having the 
:py:attr:`identifier` ``'trig-func'``. When :py:func:`export_file` is called later with the export 
file format ``'txt/table'``, the table identifier must be specified as extra argument ``key`` to select the right data table to export.
Furthermore, the table's identifier may be used as lookup key to retrieve the table from the :py:attr:`tables` dictionary view
of a :py:class:`DataCollection`."""

    class PlotMode(enum.Enum):
        """AUTODOC_SKIP_MEMBER

Members:

  NoPlot

  Line

  Histogram

  BarChart

  Scatter"""
        NoPlot = enum.auto()
        Line = enum.auto()
        Histogram = enum.auto()
        BarChart = enum.auto()
        Scatter = enum.auto()

    @property
    def axis_label_x(self) -> str:
        """The text label of the x-axis. This string is only used for a data plot if the :py:attr:`x` property of the data table is ``None`` and the x-coordinates of the data points are implicitly defined by the table's :py:attr:`interval` property. Otherwise the :py:attr:`name` of the :py:attr:`x` property is used as axis label."""
        ...

    @property
    def interval(self) -> Tuple[float, float]:
        """A pair of float values specifying the x-axis interval covered by the data points in this table.
This interval is only used by the table if the data points do not possess explicit x-coordinates (i.e. if the table's :py:attr:`x` property is ``None``).
In the absence of explicit x-coordinates, the interval specifies the range of equispaced x-coordinates implicitly generated
by the data table.

Implicit x-coordinates are typically used in data tables representing histograms, which consist of equally-sized bins
covering a certain value range along the x-axis. The bin size is then given by the interval width divided by the
number of data points (see :py:attr:`PropertyContainer.count` property). The implicit x-coordinates of data points are placed in the centers of the bins. 
You can call the table's :py:meth:`xy` method to let it explicitly calculate the x-coordinates from the value interval for every data point. 

:Default: ``(0.0, 0.0)``"""
        ...

    @property
    def plot_mode(self) -> PlotMode:
        """The type of graphical plot for rendering the data in this :py:class:`DataTable`. Must be one of the following predefined constants:

  * ``DataTable.PlotMode.NoPlot``
  * ``DataTable.PlotMode.Line``
  * ``DataTable.PlotMode.Histogram``
  * ``DataTable.PlotMode.BarChart``
  * ``DataTable.PlotMode.Scatter``


:Default: ``DataTable.PlotMode.Line``"""
        ...

    @property
    def x(self) -> Optional[Property]:
        """The :py:class:`Property` containing the x-coordinates of the data points (for the purpose of plotting). The data points may not have explicit x-coordinates, so this property may be ``None`` for a data table. In such a case, the x-coordinates of the data points are implicitly determined by the table's :py:attr:`interval`."""
        ...

    @property
    def y(self) -> Optional[Property]:
        """The :py:class:`Property` containing the y-coordinates of the data points (for the purpose of plotting). This may be a vector property having more than one component per data point, in which case this data table represents a family of data plots."""
        ...

    def xy(self) -> numpy.typing.NDArray:
        """This convenience method returns a two-dimensional NumPy array containing both the x- and the y-coordinates of the data points in this data table.
If the table has no explicit :py:attr:`x` coordinate property set, this method automatically computes equispaced x-coordinates from the :py:attr:`interval`."""
        ...

class VoxelGrid(PropertyContainer):
    """Base: :py:class:`ovito.data.PropertyContainer`

A two- or three-dimensional structured grid. Each cell (voxel) of the grid is of the same size and shape. 
The geometry of the entire grid, its :py:attr:`domain`, is defined by an attached 
:py:class:`SimulationCell` object, which specific a three-dimensional parallelepiped 
or a two-dimensional parallelogram. 
See also the corresponding user manual page for more information on this object type.

The :py:attr:`shape` property of the grid specifies the number of voxels along each 
domain cell vector. The size of an individual voxel is given by domain cell size 
divided by the number of voxels in each spatial direction. 

Every voxel of the grid may be associated with one or more field values. The data for these *voxel properties* 
is stored in standard :py:class:`Property` objects, similar to particle or bond properties. Voxel properties can be accessed by name through 
the dictionary interface that the :py:class:`VoxelGrid` class inherits from its :py:class:`PropertyContainer` 
base class. 

Voxel grids can be loaded from input data files, e.g. a CHGCAR file containing the electron density computed by the VASP code, 
or they can be dynamically generated within OVITO. The :py:class:`SpatialBinningModifier` lets you 
project the information associated with the unstructured particle set to a structured voxel grid. 

Given a voxel grid, the :py:class:`CreateIsosurfaceModifier` can then generate a :py:class:`SurfaceMesh` 
representing an isosurface for a field quantity defined on the voxel grid. 

Example

The following code example demonstrates how to create a new :py:class:`VoxelGrid` from scratch 
and initialize it with data from a Numpy array:

  ```python
  # Starting with an empty DataCollection: 
  data = DataCollection()
  
  # Create a new SimulationCell object defining the outer spatial dimensions
  # of the grid and the boundary conditions, and add it to the DataCollection: 
  data.cell = SimulationCell(pbc=(True, True, True), vis=SimulationCellVis(line_width=0.03))
  data.cell_[:,:3] = [[10,0,0],[0,10,0],[0,0,10]]
  
  # Generate a three-dimensional Numpy array containing the grid cell values.
  nx = 10; ny = 6; nz = 8
  field_data = numpy.random.random((nx, ny, nz))
  
  # Create the VoxelGrid object and give it a unique identifier by which it can be referred to later on.
  # Link the voxel grid to the SimulationCell object created above, which defines its spatial extensions.
  # Specify the shape of the grid, i.e. the number of cells in each spatial direction.
  # Finally, assign a VoxelGridVis visual element to the data object to make the grid visible in the scene.
  grid = VoxelGrid(
      identifier = 'field', 
      domain = data.cell, 
      shape = field_data.shape, 
      vis = VoxelGridVis(enabled=True, transparency=0.6))
  
  # Associate a new property with the voxel grid cells and initialize it with the data from the Numpy array. 
  # Note that the data must be provided as linear (1-dim.) array with the following type of memory layout:
  # The first grid dimension (x) is the fasted changing index while the third grid dimension (z) is the slowest varying index.
  # In this example, this corresponds to the "Fortran" memory layout of Numpy. 
  grid.create_property('Field Value', data=field_data.flatten(order='F'))
  
  # Insert the VoxelGrid object into the DataCollection.
  data.objects.append(grid)
          
  # For demonstration purposes, compute an isosurface on the basis of the VoxelGrid created above.
  data.apply(CreateIsosurfaceModifier(operate_on='voxels:field', property='Field Value', isolevel=0.7))
```"""

    @property
    def domain(self) -> Optional[SimulationCell]:
        """The :py:class:`SimulationCell` describing the (possibly periodic) domain which this grid is embedded in. Note that this cell generally is independent of and may be different from the :py:attr:`cell` found in the :py:class:`DataCollection`."""
        ...

    @property
    def shape(self) -> Tuple[int, int, int]:
        """A tuple with the numbers of grid cells along each of the three cell vectors of the :py:attr:`domain`. 

For two-dimensional grids, for which the :py:attr:`is2D` property of the :py:attr:`domain` is set to true, the third entry of the :py:attr:`shape` tuple is always equal to 1. 

Assigning new shape dimensions to the grid automatically resizes the one-dimensional data arrays of the :py:class:`PropertyContainer` base class if necessary and updates its :py:attr:`PropertyContainer.count` property to be equal to the product of the three shape dimensions, i.e. the total number of voxel cells."""
        ...

    @property
    def vis(self) -> Optional[ovito.vis.VoxelGridVis]:
        """The :py:class:`DataVis` element currently associated with this data object, which is responsible for visually rendering the stored data. If set to ``None``, the data object remains non-visual and doesn't appear in rendered images or the viewports. Furthermore, note that the same :py:class:`DataVis` element may be assigned to multiple data objects in order to synchronize their visual appearance."""
        ...